"""
Entry point to working with licensing on Portal or ArcGIS Online
"""
from ..._impl.connection import _ArcGISConnection
from ..._impl.common._mixins import PropertyMap
from ...gis import GIS
from ._base import BasePortalAdmin
########################################################################
class LicenseManager(BasePortalAdmin):
    """
    Provides tools to work and manage licenses in ArcGIS Online and
    ArcGIS Enterprise (Portal)

    ===============     ====================================================
    **Argument**        **Description**
    ---------------     ----------------------------------------------------
    url                 required string, the web address of the site to
                        manage licenses.
                        example:
                        https://<org url>/<wa>/sharing/rest/portals/self/purchases
    ---------------     ----------------------------------------------------
    gis                 required GIS, the gis connection object
    ===============     ====================================================

    :returns:
       LicenseManager Object
    """
    _con = None
    _url = None
    _json_dict = None
    _json = None
    _properties = None
    def __init__(self, url, gis=None, initialize=True, **kwargs):
        """class initializer"""
        super(LicenseManager, self).__init__(url=url, gis=gis)
        self._url = url
        if isinstance(gis, _ArcGISConnection):
            self._con = gis
        elif isinstance(gis, GIS):
            self._gis = gis
            self._con = gis._con
        else:
            raise ValueError(
                "connection must be of type GIS or _ArcGISConnection")
        if initialize:
            self._init(connection=self._con)
    #----------------------------------------------------------------------
    def get(self, name):
        """
        retrieves a license by it's name (title)
        ===============     ====================================================
        **Argument**        **Description**
        ---------------     ----------------------------------------------------
        name                required string, name of the entitlement to locate
                            on the organization.
                            example:
                            name="arcgis pro"
        ===============     ====================================================

        :returns:
           License Object
        """
        licenses = self.all()
        for l in licenses:
            if 'listing' in l.properties and \
               'title' in l.properties['listing'] and \
                l.properties['listing']['title'].lower() == name.lower():
                return l
            del l
        del licenses
        return None
    #----------------------------------------------------------------------
    def all(self):
        """
        Returns all Licenses registered with an organization

        :returns:
           list of License objects
        """
        licenses = []
        if self._properties is None:
            self._init()
        if 'purchases' in self.properties:
            purchases = self.properties['purchases']
            for purchase in purchases:
                licenses.append(License(gis=self._gis, info=purchase))
        return licenses
    #----------------------------------------------------------------------
    @property
    def bundles(self):
        """
        Returns a list of Application Bundles for an Organization

        :returns:
           list of Bundle objects

        """
        if self._gis.version < [6,4]:
            raise NotImplementedError("`bundles` not implemented before version 6.4")

        url = "{base}/portals/self/appBundles".format(base=self._gis._portal.resturl)
        params = {
            'f' : 'json',
            'num' : 100,
            'start' : 1
        }

        res = self._con.get(url, params)
        buns = res['appBundles']
        while res['nextStart'] != -1:
            params['start'] = res['nextStart']
            res = self._gis._con.get(url, params)
            buns += res['appBundles']
            if res['nextStart'] == -1:
                break
        return [Bundle(url="{base}content/listings/{id}".format(base=self._gis._portal.resturl,
                                                id=b["appBundleItemId"]),
                       properties=b,
                       gis=self._gis)
                for b in buns]
    #----------------------------------------------------------------------
    @property
    def offline_pro(self):
        """
        Administrators can get/set the disconnect settings for the ArcGIS Pro licensing.
        A value of True means that a user can check out a license from the enterprise
        inorder to use it in a disconnected setting.  By setting `offline_pro` to False,
        the enterprise users cannot check out licenses to work in a disconnected setting
        for ArcGIS Pro.

        :returns: Boolean

        """
        lic = self.get("arcgis pro")
        return lic.properties.provision.canDisconnect
    #----------------------------------------------------------------------
    @offline_pro.setter
    def offline_pro(self, value):
        """
        Administrators can get/set the disconnect settings for the ArcGIS Pro licensing.
        A value of True means that a user can check out a license from the enterprise
        inorder to use it in a disconnected setting.  By setting `offline_pro` to False,
        the enterprise users cannot check out licenses to work in a disconnected setting
        for ArcGIS Pro.

        """
        import json
        lic = self.get("arcgis pro")
        url = "{base}content/listings/{itemid}/setDisconnectSettings".format(
            base=self._gis._portal.resturl,
            itemid=lic.properties.provision.itemId
        )
        params = {
            "f" : "json",
            "canDisconnect" : json.dumps(value),
            "maxDisconnectDuration": -1
        }
        res = self._con.post(url, params)
        if 'success' in res and \
           res['success'] == False:
            raise Exception("Could not update the dicconnect settings.")
        elif 'success' not in res:
            raise Exception("Could not update the dicconnect settings: %s" % res)


########################################################################
class Bundle(object):
    """
    This represents a single instance of an application bundle
    """
    _users = None
    _con = None
    _url = None
    _properties = None
    _gis = None
    _id = None
    #----------------------------------------------------------------------
    def __init__(self, url, properties=None, gis=None):
        """Constructor"""
        import os
        if gis is None:
            import arcgis
            gis = arcgis.env.active_gis
        self._gis = gis
        self._url = url
        self._con = gis._con
        self._id = os.path.basename(url)
        self._properties = properties
    #----------------------------------------------------------------------
    def _find(self, appid):
        """if properties are missing, populate the app bundle's properties"""
        url = "{base}/portals/self/appBundles".format(base=self._gis._portal.resturl)
        params = {
            'f' : 'json',
            'num' : 100,
            'start' : 1
        }

        res = self._con.get(url, params)
        buns = res['appBundles']
        while res['nextStart'] != -1:
            params['start'] = res['nextStart']
            res = self._gis._con.get(url, params)
            buns += res['appBundles']
            if res['nextStart'] == -1:
                break
        for b in buns:
            if b['appBundleItemId'] == self._id:
                return b
        raise ValueError("Invalid Application Bundle.")
    #----------------------------------------------------------------------
    @property
    def properties(self):
        """returns the application bundles properties"""
        if self._properties is None:
            try:
                params = {'f': 'json'}
                r = self._con.get(self._url, params)
                self._properties = PropertyMap(r)
            except:
                self._properties = PropertyMap({})
        elif isinstance(self._properties, dict):
            self._properties = PropertyMap(self._properties)
        return self._properties
    #----------------------------------------------------------------------
    @property
    def users(self):
        """returns a list of users assigned the application bundle"""
        if self._users is None:
            self._users = self._gis.users.search("appbundle:%s" % self.properties.appBundleItemId)
        return self._users
    #----------------------------------------------------------------------
    def __len__(self):
        return len(self.users)
    #----------------------------------------------------------------------
    def __str__(self):
        """"""
        return "<AppBundle: %s >" % self.properties['name']
    #----------------------------------------------------------------------
    def __repr__(self):
        """"""
        return "<AppBundle: %s >" % self.properties['name']
    #----------------------------------------------------------------------
    def assign(self, users):
        """
        Assigns the current application bundle to a list of users

        ===============     ====================================================
        **Argument**        **Description**
        ---------------     ----------------------------------------------------
        users               Required List. A list of user names or User objects
                            to assign the current application bundle to.
        ===============     ====================================================


        :returns: boolean

        """
        if isinstance(users, (tuple, set, list)) == False:
            users = [users]
        from arcgis.gis import User
        url = "%s%s" % (self._url, "/provisionUserAppBundle")
        params = {
            'f' : 'json',
            'users' : None,
            'revoke' : False
        }
        us = []
        for user in users:
            if isinstance(users, str):
                us.append(user)
            elif isinstance(user, User):
                us.append(user.username)
        params['users'] = ",".join(us)
        res = self._con.post(url, params)
        self._users = None
        self._properties = None
        if 'success' in res:
            return res['success']
        return res
    #----------------------------------------------------------------------
    def revoke(self, users):
        """
        Revokes the current application bundle to a list of users

        ===============     ====================================================
        **Argument**        **Description**
        ---------------     ----------------------------------------------------
        users               Required List. A list of user names or User objects
                            to remove the current application bundle to.
        ===============     ====================================================


        :returns: boolean

        """
        if isinstance(users, (tuple, set, list)) == False:
            users = [users]
        from arcgis.gis import User
        url = "%s%s" % (self._url, "/provisionUserAppBundle")
        params = {
            'f' : 'json',
            'users' : None,
            'revoke' : True
        }
        us = []
        self._users = None
        self._properties = None
        for user in users:
            if isinstance(users, str):
                us.append(user)
            elif isinstance(user, User):
                us.append(user.username)
        params['users'] = ",".join(us)
        res = self._con.post(url, params)
        if 'success' in res:
            return res['success']
        return res


########################################################################
class License(object):
    """
    Represents a single entitlement for a given organization.


    ===============     ====================================================
    **Argument**        **Description**
    ---------------     ----------------------------------------------------
    gis                 required GIS, the gis connection object
    ---------------     ----------------------------------------------------
    info                required dictionary, the information provided by
                        the organization's site containing the provision
                        and listing information.
    ===============     ====================================================

    :returns:
       License Object
    """
    _properties = None
    _gis = None
    _con = None
    #----------------------------------------------------------------------
    def __init__(self, gis, info):
        """Constructor"""
        self._gis = gis
        self._con = gis._con
        self._properties = PropertyMap(info)
    #----------------------------------------------------------------------
    def __str__(self):
        try:
            return '<%s %s at %s>' % (self.properties['listing']['title'],type(self).__name__, self._gis._portal.resturl)
        except:
            return '<%s at %s>' % (type(self).__name__, self._gis._portal.resturl)
    #----------------------------------------------------------------------
    def __repr__(self):
        try:
            return '<%s %s at %s>' % (self.properties['listing']['title'],type(self).__name__, self._gis._portal.resturl)
        except:
            return '<%s at %s>' % (type(self).__name__, self._gis._portal.resturl)
    #----------------------------------------------------------------------
    @property
    def properties(self):
        return self._properties
    #----------------------------------------------------------------------
    @property
    def report(self):
        """
        returns a Panda's Dataframe of the licensing count.
        """
        import pandas as pd
        data = []
        columns = ['Entitlement', 'Total', 'Assigned', 'Remaining']
        if 'provision' in self.properties:
            for k,v in self.properties['provision']['orgEntitlements']['entitlements'].items():
                counter = 0
                for u in self.all():
                    if k in u['entitlements']:
                        counter += 1
                row = [k, v['num'], counter, v['num'] - counter]
                data.append(row)
                del k, v
        return pd.DataFrame(data=data, columns=columns)
    #----------------------------------------------------------------------
    def plot(self):
        """returns a simple bar chart of assigned and remaining entitlements"""
        report = self.report
        return report.plot(x=report["Entitlement"],
                           y=['Assigned', 'Remaining'],
                           kind='bar',stacked=True)
    #----------------------------------------------------------------------
    def all(self):
        """
        returns a list of all usernames and their entitlements for this license
        """
        item_id = self.properties['listing']['itemId']
        url = "%scontent/listings/%s/userEntitlements" % (self._gis._portal.resturl, item_id)
        start = 1
        num = 100
        params = {
            'start' : start,
            'num' : num
        }
        user_entitlements = []
        res = self._con.get(url, params)
        user_entitlements += res['userEntitlements']
        if 'nextStart' in res:
            while res['nextStart'] > 0:
                start += num
                params = {
                    'start' : start,
                    'num' : num
                }
                res = self._con.get(url, params)
                user_entitlements += res['userEntitlements']
        return user_entitlements
    #----------------------------------------------------------------------
    def user_entitlement(self, username):
        """
        checks if a user has the entitlement assigned to them

        ===============     ====================================================
        **Argument**        **Description**
        ---------------     ----------------------------------------------------
        username            required string, the name of the user you want to
                            examine the entitlements for.
        ===============     ====================================================

        :returns:
           dictionary
        """
        item_id = self.properties['listing']['itemId']
        url = "%scontent/listings/%s/userEntitlements" % (
            self._gis._portal.resturl,
            item_id)
        start = 1
        num = 100
        params = {
            'start' : start,
            'num' : num
        }
        user_entitlements = []
        res = self._con.get(url, params)
        for u in res['userEntitlements']:
            if u['username'].lower() == username.lower():
                return u
        if 'nextStart' in res:
            while res['nextStart'] > 0:
                start += num
                params = {
                    'start' : start,
                    'num' : num
                }
                res = self._con.get(url, params)
                for u in res['userEntitlements']:
                    if u['username'].lower() == username.lower():
                        return u
        return {}
    #----------------------------------------------------------------------
    def assign(self, username, entitlements, suppress_email=True):
        """
        grants a user an entitlement.
        ===============     ====================================================
        **Argument**        **Description**
        ---------------     ----------------------------------------------------
        username            required string, the name of the user you wish to
                            assign an entitlement to.
        ---------------     ----------------------------------------------------
        entitlments         required list, a list of entitlements values
        ---------------     ----------------------------------------------------
        suppress_email       optional boolean, if True, the org will not notify
                            a user that their entitlements has changed (default)
                            If False, the org will send an email notifying a
                            user that their entitlements have changed.
        ===============     ====================================================

        :returns:
           boolean
        """
        item_id = self.properties['listing']['itemId']
        if isinstance(entitlements, str):
            entitlements = entitlements.split(',')
        params = {
            "f" : "json",
            "userEntitlements" : {"users":[username],
                                  "entitlements":entitlements},

        }
        if suppress_email is not None:
            params["suppressCustomerEmail"] = suppress_email
        url = "%scontent/listings/%s/provisionUserEntitlements" % (self._gis._portal.resturl, item_id)
        res = self._con.post(url, params)
        if 'success' in res:
            return res['success'] == True
        return res
    #----------------------------------------------------------------------
    def revoke(self, username, entitlements, suppress_email=True):
        """
        removes a specific license from a given entitlement

        ===============     ====================================================
        **Argument**        **Description**
        ---------------     ----------------------------------------------------
        username            required string, the name of the user you wish to
                            assign an entitlement to.
        ---------------     ----------------------------------------------------
        entitlments         required list, a list of entitlements values,
                            if * is given, all entitlements will be revoked
        ---------------     ----------------------------------------------------
        suppress_email      optional boolean, if True, the org will not notify
                            a user that their entitlements has changed (default)
                            If False, the org will send an email notifying a
                            user that their entitlements have changed.
        ===============     ====================================================

        :returns:
           boolean
        """
        if entitlements == "*":
            return self.assign(username=username,
                                  entitlements=[],
                                  suppress_email=suppress_email)
        elif isinstance(entitlements, list):
            es = self.user_entitlement(username=username)

            if 'entitlements' in es:
                lookup = {e.lower() : e for e in es['entitlements']}
                es = [e.lower() for e in es['entitlements']]
                if isinstance(entitlements, str):
                    entitlements = [entitlements]
                entitlements = list(set(es) - set([e.lower() for e in entitlements]))
                es2 = []
                for e in entitlements:
                    es2.append(lookup[e])
                return self.assign(username=username,
                                   entitlements=es2,
                                   suppress_email=suppress_email)
        return False



""" Defines store functions for working with assignments.
"""

from ... import workforce
from .dispatchers import query_dispatchers
from .utils import add_features, remove_features, update_features, validate


def get_assignment(project, object_id=None, global_id=None):
    """ Gets the identified Assignment.  Exactly one form of identification should be provided.
        :param project:
        :param object_id: The assignment's OBJECTID.
        :param global_id: The assignment's GlobalID.
        :returns: Assignment
    """
    if object_id:
        where = "{} = {}".format(project._assignment_schema.object_id, object_id)
    elif global_id:
        where = "{} = '{}'".format(project._assignment_schema.global_id, global_id)
    else:
        where = "1=0"
    assignments = query_assignments(project, where)
    if assignments:
        return assignments[0]
    return None


def get_assignments(project):
    """ Gets all Assignments in the project.
        :param project:
        :returns: list of Assignments
    """
    return query_assignments(project, '1=1')


def query_assignments(project, where='1=1'):
    """ Executes a query against the assignments feature layer.
        :param project: The project in which to query assignments.
        :param where: An ArcGIS where clause.
        :returns: list of Assignments
    """
    assignments = []
    assignment_features = project.assignments_layer.query(where, return_all_freatures=True).features
    # fetch assignment types, dispatchers, and workers
    project._update_cached_objects()
    # refresh cached objects
    for feature in assignment_features:
        assignments.append(workforce.Assignment(project, feature))
    return assignments


def add_assignments(project, assignments):
    """ Adds Assignments to a project.

        Side effect: Upon successful addition on the server, the object_id and global_id fields of
        each Assignment in assignments will be updated to the values assigned by the server.

        :param project:
        :param assignments: list of Assignments
        :returns the list of Assignments
        :raises ValidationError: Indicates that one or more assignments failed validation.
        :raises ServerError: Indicates that the server rejected the assignments.
    """
    project._update_cached_objects()
    use_global_ids = True
    for assignment in assignments:
        assignment.project = project
        validate(assignment._validate_for_add)
        if assignment.global_id is None:
            use_global_ids = False

    features = [assignment.feature for assignment in assignments]
    add_features(project.assignments_layer, features, use_global_ids)
    return assignments


def add_assignment(project, feature=None, geometry=None, assignment_type=None,
            assigned_date=None, assignment_read=None, completed_date=None, declined_comment=None,
            declined_date=None, description=None, dispatcher=None, due_date=None, in_progress_date=None,
            location=None, notes=None, paused_date=None, priority=None, status=None,
            work_order_id=None, worker=None):
    """
    Adds a new assignment to the project
    """
    project._update_cached_objects()
    assignment = workforce.Assignment(project,
                                      feature,
                                      geometry,
                                      assignment_type,
                                      assigned_date,
                                      assignment_read,
                                      completed_date,
                                      declined_comment,
                                      declined_date,
                                      description,
                                      dispatcher,
                                      due_date,
                                      in_progress_date,
                                      location,
                                      notes,
                                      paused_date,
                                      priority,
                                      status,
                                      work_order_id,
                                      worker
                                      )

    return add_assignments(project, [assignment])[0]


def update_assignments(project, assignments):
    """ Updates Assignments.
        :param project:
        :param assignments: list of Assignments to update
        :raises ValidationError: Indicates that one or more assignments failed validation.
        :raises ServerError: Indicates that the server rejected the updates.
    """
    project._update_cached_objects()
    for assignment in assignments:
        validate(assignment._validate_for_update)
    features = [assignment.feature for assignment in assignments]
    update_features(project.assignments_layer, features)
    return assignments
    
    
def update_assignment(project, assignment, geometry=None, assignment_type=None,
            assigned_date=None, assignment_read=None, completed_date=None, declined_comment=None,
            declined_date=None, description=None, dispatcher=None, due_date=None, in_progress_date=None,
            location=None, notes=None, paused_date=None, priority=None, status=None,
            work_order_id=None, worker=None):
    """
    Sets the properties of an assignment and updates the item on the server
    """
    project._update_cached_objects()
    if geometry:
        assignment.geometry = geometry
    if assigned_date:
        assignment.assigned_date = assigned_date
    if assignment_read:
        assignment.assignment_read = assignment_read
    if completed_date:
        assignment.completed_date = completed_date
    if declined_comment:
        assignment.declined_comment = declined_comment
    if declined_date:
        assignment.declined_date = declined_date
    if description:
        assignment.description = description
    if due_date:
        assignment.due_date = due_date
    if in_progress_date:
        assignment.in_progress_date = in_progress_date
    if location:
        assignment.location = location
    if notes:
        assignment.notes = notes
    if paused_date:
        assignment.paused_date = paused_date
    if priority:
        assignment._priority = priority
    if status:
        assignment.status = status
    if work_order_id:
        assignment.work_order_id = work_order_id
    if assignment_type:
        assignment.assignment_type = assignment_type
    if dispatcher:
        assignment.dispatcher = dispatcher
    if worker:
        assignment.worker = worker
    return update_assignments(project, [assignment])[0]


def delete_assignments(project, assignments):
    """ Removes assignments from the project.
        :param project:
        :param assignments: list of Assignments
        :raises ValidationError: Indicates that one or more assignments failed validation.
        :raises ServerError: Indicates that the server rejected the removal.
    """
    project._update_cached_objects()
    for assignment in assignments:
        validate(assignment._validate_for_remove)
    features = [assignment.feature for assignment in assignments]
    remove_features(project.assignments_layer, features)


def dispatchers_for_assignment_features(project, features):
    dispatcher_ids = [feature.attributes.get(project._assignment_schema.dispatcher_id)
                      for feature in features]
    dispatcher_ids = [did for did in dispatcher_ids if did is not None]
    if dispatcher_ids:
        dispatchers_where = "{} IN ({})".format(project._dispatcher_schema.object_id,
                                                ','.join(map(str, dispatcher_ids)))
        return query_dispatchers(project, dispatchers_where)
    return []


def workers_for_assignment_features(project, features):
    worker_ids = [feature.attributes.get(project._assignment_schema.worker_id)
                  for feature in features]
    worker_ids = [wid for wid in worker_ids if wid is not None]
    if worker_ids:
        workers_where = "{} IN ({})".format(project._worker_schema.object_id,
                                            ','.join(map(str, worker_ids)))
        return workforce._store.query_workers(project, workers_where)
    return []



'''
The orthomapping python API allows automating orthomapping tasks in the server environment.

For more information about orthomapping workflows in ArcGIS, please visit the help documentation at 
http://desktop.arcgis.com/en/arcmap/10.4/manage-data/raster-and-images/block-adjustment-for-mosaic-datasets.htm.
'''


import arcgis
import json
import string as _string
import random as _random
from arcgis.gis import Item
import collections
from ._util import _set_context

from arcgis.geoprocessing._support import _analysis_job, _analysis_job_results, \
                                          _analysis_job_status, _layer_input

###################################################################################################
###
### INTERNAL FUNCTIONS
###
###################################################################################################

def _execute_task(gis, taskname, params):

    gptool_url = gis.properties.helperServices.orthoMapping.url
    gptool = arcgis.gis._GISResource(gptool_url, gis)
    task = taskname

    task_url, job_info, job_id = _analysis_job(gptool, task, params)
    #print ('task url is ', task_url)

    job_info = _analysis_job_status(gptool, task_url, job_info)
    job_values = _analysis_job_results(gptool, task_url, job_info, job_id)

    item_properties = {
        "properties":{
            "jobUrl": task_url + '/jobs/' + job_info['jobId'],
            "jobType": "GPServer",
            "jobId": job_info['jobId'],
            "jobStatus": "completed"
            }
        }
    return job_values


#def _id_generator(size=6, chars=_string.ascii_uppercase + _string.digits):
    #return ''.join(_random.choice(chars) for _ in range(size))
###################################################################################################
###################################################################################################
def _set_image_collection_param(gis, params, image_collection):
    if isinstance(image_collection, str):
        if 'http:' in image_collection or 'https:' in image_collection:
            params['imageCollection'] = json.dumps({ 'url' : image_collection })
        else:
            params['imageCollection'] = json.dumps({ 'uri' : image_collection })
    elif isinstance(image_collection, Item):
        params['imageCollection'] = json.dumps({ "itemId" : image_collection.itemid })
    else:
        raise TypeError("image_collection should be a string (service name) or Item")

    return


def _create_output_image_service(gis, output_name, task):
    ok = gis.content.is_service_name_available(output_name, "Image Service")
    if not ok:
        raise RuntimeError("An Image Service by this name already exists: " + output_name)

    create_parameters = {
        "name": output_name,
        "description": "",
        "capabilities": "Image",
        "properties": {
            "path": "@",
            "description": "",
            "copyright": ""
        }
    }

    output_service = gis.content.create_service(output_name, create_params=create_parameters,
                                                      service_type="imageService")
    description = "Image Service generated from running the " + task + " tool."
    item_properties = {
        "description": description,
        "tags": "Analysis Result, " + task,
        "snippet": "Analysis Image Service generated from " + task
    }
    output_service.update(item_properties)
    return output_service

###################################################################################################
###
### PUBLIC API 
###
###################################################################################################
def is_supported(gis=None):
    """
    Returns True if the GIS supports orthomapping. If a gis isn't specified,
    checks if arcgis.env.active_gis supports raster analytics
    """
    gis = arcgis.env.active_gis if gis is None else gis
    if 'orthoMapping' in gis.properties.helperServices:
        return True
    else:
        return False

###################################################################################################
## Compute Sensor model
###################################################################################################
def compute_sensor_model(image_collection, 
                         mode='QUICK', 
                         location_accuracy='High', 
                         context=None,
                         gis = None):
    """
    compute_sensor_model computes the bundle block adjustment for the image collection 
    and applies the frame xform to the images. It will also generate the control point 
    table, solution table, solution points table and flight path table. 
    These tables will not be published as Portal items. 

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    image_collection       Required, the input image collection on which to compute
                           the sensor model.
                           The image_collection can be a portal Item or an image service URL or a URI
                           
                           The image_collection must exist.
    ------------------     --------------------------------------------------------------------
    mode                   Optional string.  the mode to be used for bundle block adjustment
                           Only the following modes are supported:
                           
                           - 'Quick' : Computes tie points and adjustment at 8x of the source imagery resolution

                           - 'Full'  : adjust the images in Quick mode then at 1x of the source imagery resolution

                           - 'Refine' : adjust the image at 1x of the source imagery resolution

                           By default, 'Quick' mode is applied to compute the sensor model.
    ------------------     --------------------------------------------------------------------
    location_acuracy       Optional string. this option allows users to specify the GPS location accuracy level of the  
                           source image. It determines how far the underline tool will search for neighboring 
                           matching images, then calculate tie points and compute adjustments.

                           Possible values for location_accuracy are:

                           - 'High'    : GPS accuracy is 0 to 10 meters, and the tool uses a maximum of 4 by 3 images 

                           - 'Medium'  : GPS accuracy of 10 to 20 meters, and the tool uses a maximum of 4 by 6 images

                           - 'Low'     : GPS accuracy of 20 to 50 meters, and the tool uses a maximum of 4 by 12 images 

                           - 'Verylow' : GPS accuracy is more than 50 meters, and the tool uses a maximum of 4 by 20 images

                           The default location_accuracy is 'High' 
    ------------------     --------------------------------------------------------------------
    context                Optional dictionary. The context parameter is used to configure additional client settings 
                           for block adjustment. The supported configurable parameters are for compute mosaic dataset
                           candidates after the adjustment. 

                           Example: 
                           {
                           "computeCandidate": False,
                           "maxoverlap": 0.6,
                           "maxloss": 0.05,
                           }
    ------------------     --------------------------------------------------------------------
    gis                    Optional GIS. The GIS on which this tool runs. If not specified, the active GIS is used.
    ==================     ====================================================================

    :return:
        The imagery layer url

    """
    gis = arcgis.env.active_gis if gis is None else gis

    params = {}

    _set_image_collection_param(gis, params, image_collection)
    modeAllowedValues = ["FULL","QUICK","REFINE"]
    if not mode.upper() in modeAllowedValues:
        raise RuntimeError("mode can only be one of the following: Quick, Full or Refine")

    params['mode'] = mode

    locationAccAllowedValues = ['HIGH', 'MEDIUM', 'LOW', 'VERYLOW']
    if not location_accuracy.upper() in locationAccAllowedValues:
        raise RuntimeError('location_accuracy can only be one of the following: '+ str(locationAccAllowedValues))

    params['locationAccuracy'] = location_accuracy

    _set_context(params, context)

    task = 'ComputeSensorModel'
    job_values = _execute_task(gis, task, params)
    
    return job_values["result"]["url"]

###################################################################################################
## Alter processing states
###################################################################################################
def alter_processing_states(image_collection, new_states, gis = None):
    '''
    Alter the processing states of the image collection.
    The states are stored as key property "Orthomapping". 
    The content of the state is a dictionary including 
    several properties which can be set based on the process 
    done on the image collection. 

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    image_collection       Required, This is the image collection that will be adjusted.

                           The image_collection can be a portal Item or an image service URL or URI
                           
                           The image_collection must exist.
    ------------------     --------------------------------------------------------------------
    new_states             Required dictionary. The state to set on the image_collection

                           This a dictionary of states that should be set on the image collection
                           The new states that can be set on the image collection are:
                           blockadjustment, dem, gcp, seamlines, colorcorrection, adjust_index, imagetype

                           Example:
                           {"blockadjustment": "raw",
                            "dem": "Dense_Natual_Neighbor",
                            "seamlines":"VORONOI",
                            "colorcorrection":"SingleColor",
                            "imagetype": "UAV/UAS",
                            "adjust_index": 0}
    ------------------     --------------------------------------------------------------------
    gis                    Optional GIS. The GIS on which this tool runs. If not specified, the active GIS is used.
    ==================     ====================================================================

    :return:
        The result will be the newly set states dictionary

    '''

    gis = arcgis.env.active_gis if gis is None else gis
        
    params = {}

    _set_image_collection_param(gis, params, image_collection)

    newStatesAllowedValues = ['blockadjustment', 'dem', 'gcp', 'seamlines', 'colorcorrection', 'adjust_index', 'imagetype']

    for key in new_states:
        if not key in newStatesAllowedValues:
            raise RuntimeError('new_states can only be one of the following: ' + str(newStatesAllowedValues))

    params['newStates'] = json.dumps(new_states)
    task = 'AlterProcessingStates'
    job_values = _execute_task(gis, task, params)
    if "processingStates" in job_values:
        if isinstance(job_values["processingStates"], dict):
            return job_values["processingStates"]
        elif isinstance(job_values["processingStates"], str):
            processing_states = job_values['processingStates'].replace("'",'"')
            processing_states=json.loads( processing_states.replace('u"','"'))
            return processing_states
 

###################################################################################################
## Get processing states
###################################################################################################
def get_processing_states(image_collection, gis = None):
    '''
    Retrieve the processing states of the image collection

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    image_collection       Required, This is the image collection that will be adjusted.

                           The image_collection can be a portal Item or an image service URL or URI
                            
                           The image_collection must exist.
    ------------------     --------------------------------------------------------------------
    gis                    Optional GIS. The GIS on which this tool runs. If not specified, the active GIS is used.
    ==================     ====================================================================

    :return:
        The result will be the newly set states dictionary

    '''

    gis = arcgis.env.active_gis if gis is None else gis

    params = {}

    _set_image_collection_param(gis, params, image_collection)

    task = 'GetProcessingStates'
    job_values = _execute_task(gis, task, params)
    return job_values["processingStates"]

"""
###################################################################################################
## Append control points
###################################################################################################
def append_control_points(image_collection, control_points, gis = None):
    '''
    Append additional ground control point sets to the image collection's control points. 
    A complete ground control point (GCP) set should have one ground control point 
    and multiple (more than 3) tie points.

    See http://desktop.arcgis.com/en/arcmap/10.3/manage-data/raster-and-images/block-adjustment-for-mosaic-datasets.htm#ESRI_SECTION1_6676F2BB9A6B453E9EE1E00B42C4A5C1
    for more information about preparing control points and tie points for orthomapping

    Parameters
    ----------
    image_collection    :   Required, the input image collection on which to compute
                            the sensor model.

                            The image_collection can be a portal Item or an image service URL
                            
                            The image_collection must exist.

    control_points      :   Required, a list of control point objects.

                            A control point object is a dictionary with key-value
                            pairs as described below:

                            The schema of control points follows the schema 
                            of the mosaic dataset control point table. The following are
                            required when defining control points:
                            - The control points must contain a Point geometry object 

                            - There must be one attribute set, describing the attributes of the control point. 

                              The control point attributes is a dictionary that must contain the following 
                              key-value pairs:

                              -- imageID (int) - Image identification using the ObjectID from the mosaic dataset footprint table.

                              -- pointID (int) - The ID of the point within the control point table

                              -- type (int)    - The type of the control point as determined by its numeric value
                                                 1: Tie Point 
                                                 2: Ground Control Point.
                                                 3: Check Point

                              -- status (int)  - The status of the point. A value of 0 indicates that the point will
                                                 not be used in computation. A non-zero value indicates otherwise.

                            Example:
                            {"geometry": {
                                "x":-118.15,"y":33.80,"z":10.0,
                                "spatialReference":{"wkid":4326}},  
                                "attributes": {
                                   "imageID": 22,
                                   "pointID": 2, 
                                   "type": 2,
                                   "status": 1, 
                                 },
                            <more points>
                            }

    '''
    gis = arcgis.env.active_gis if gis is None else gis

    params = {}

    _set_image_collection_param(params, image_collection)

    params['controlPoints'] = json.dumps(control_points)
    task = 'AppendControlPoints'
    _execute_task(gis, task, None)
    return
"""
###################################################################################################
## Match control points
###################################################################################################
def match_control_points(image_collection, control_points, similarity=None, context=None, gis = None):
    '''
    If each of the ground control point has at least one matching tie point. 
    Then, this function will help find all remaining tie points on all other image items of the 
    image collection.

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    image_collection       Required, the input image collection that will be adjusted.

                           The image_collection can be a portal Item or an image service URL or a URI
                            
                           The image_collection must exist.
    ------------------     --------------------------------------------------------------------
    control_points         Required, a list of control point sets objects.

                           The schema of control points follows the schema 
                           of the mosaic dataset control point table. 

                           The control point object should contain the point geometry, pointID, type, status and the
                           imagePoints. (the imagePoints attribute inside the control points object lists the imageIDs)

                           -- pointID (int) - The ID of the point within the control point table.

                           -- type (int)    - The type of the control point as determined by its numeric value
                                                 1: Tie Point 
                                                 2: Ground Control Point.
                                                 3: Check Point

                           -- status (int)  - The status of the point. A value of 0 indicates that the point will
                                                 not be used in computation. A non-zero value indicates otherwise.

                           

                           -- imageID (int) - Image identification using the ObjectID from the mosaic dataset footprint table.

                           Example:
                           [{
                           "status": 1,
                           "type": 2,
                           "x": -117.0926538,
                           "y": 34.00704253,
                           "z": 634.2175,
                           "spatialReference": {
                               "wkid": 4326
                           }, // default WGS84
                           "imagePointSpatialReference": {}, // default ICS
                           "pointId": 1,
                           "xyAccuracy": "0.008602325",
                           "zAccuracy": "0.015",
                           "imagePoints": [{
                               "imageID": 1,
                               "x": 2986.5435987557084,
                               "y": -2042.5193648409431,
                               "u": 3057.4580682832734,
                               "v": -1909.1506872159698
                           },
                           {
                               "imageID": 2,
                               "x": 1838.2814361401108,
                               "y": -2594.5280063817972,
                               "u": 3059.4079724863363,
                               "v": -2961.292545463305
                           },
                           {
                               "imageID": 12,
                               "x": 5332.855578204663,
                               "y": -2533.2805429751907,
                               "u": 614.2338676573158,
                               "v": -165.10836768947297
                           },
                           {
                               "imageID": 13,
                               "x": 4932.0895715254455,
                               "y": -1833.8401744114287,
                               "u": 616.9396928182223,
                               "v": -1243.1445126959693
                           }]
                           },
                           …
                           …
                           ] 
    ------------------     --------------------------------------------------------------------
    similarity             Optional string. Choose the tolerance level for your control point matching. 

                           Low- The similarity tolerance for finding control points will be low.
                           This option will produce the most control points, 
                           but some may have a higher level of error. 

                           Medium - The similarity tolerance for finding control points will be medium.
                           
                           High - The similarity tolerance for finding control points will be high. 
                           This option will produce the least number of control points, 
                           but each matching pair will have a lower level of error. This is the default. 
    ------------------     --------------------------------------------------------------------
    context                Optional dictionary.Additional settings such as the input control points 
                           spatial reference can be specified here. 

                           For example:
                           {"groundControlPointsSpatialReference": {"wkid": 3459}, "imagePointSpatialReference": {"wkid": 3459}}

                           Note: The ground control points spatial reference and image point spatial reference 
                           spatial reference set in the context parameter is to decide the returned point set's 
                           ground control points spatial reference and image point spatial reference. 
                           If these two parameters are not set here, the tool will use the spatial reference 
                           defined in the input point set. And if no spatial reference is defined in the point set,
                           then the default ground control points coordinates are in lon/lat and image points 
                           coordinates are in image coordinate system. 
    ------------------     --------------------------------------------------------------------
    gis                    Optional GIS. The GIS on which this tool runs. If not specified, the active GIS is used.
    ==================     ====================================================================

    :return:
        A dictionary object

    '''

    gis = arcgis.env.active_gis if gis is None else gis

    params = {}

    _set_image_collection_param(gis, params, image_collection)

    params['inputControlPoints'] = json.dumps(control_points)

    similarityAllowedValues = ['Low', 'Medium', 'High']
    if similarity is not None:
        if not similarity in similarityAllowedValues:
            raise RuntimeError('similarity can only be one of the following: '+str(similarityAllowedValues))
        params['similarity'] = similarity

    if context is not None:
        params['context'] = json.dumps(context)

    task = 'MatchControlPoints'
    job_values = _execute_task(gis, task, params)

    if job_values["result"] is not None:
        gptool_url = gis.properties.helperServices.orthoMapping.url
        gptool = arcgis.gis._GISResource(gptool_url, gis)
        result = gptool._con.post(job_values["result"]["url"],{},token=gptool._token)
    else:
        return job_values["result"]

    return result

###################################################################################################
## Color Correction
###################################################################################################
def color_correction(image_collection,
                  color_correction_method,
                  dodging_surface_type,
                  target_image=None,
                  context = None,
                  gis = None):
    '''
    Color balance the image collection. 
    Refer to the "Color Balance Mosaic Dataset" GP tool for 
    documentation on color balancing mosaic datasets.
    http://pro.arcgis.com/en/pro-app/tool-reference/data-management/color-balance-mosaic-dataset.htm

    ====================================     ====================================================================
    **Argument**                             **Description**
    ------------------------------------     --------------------------------------------------------------------
    image_collection                         Required. This is the image collection that will be adjusted.

                                             The image_collection can be a portal Item or an image service URL or a URI
                            
                                             The image_collection must exist.
    ------------------------------------     --------------------------------------------------------------------
    color_correction_method                  Required string. This is the method that will be used for color
                                             correction computation. The available options are:

                                             Dodging-Change each pixel's value toward a target color. 
                                             With this technique, you must also choose 
                                             the type of target color surface, which 
                                             affects the target color. Dodging tends 
                                             to give the best result in most cases. 

                                             Histogram-Change each pixel's value according
                                             to its relationship with a target histogram. 
                                             The target histogram can be derived from
                                             all of the rasters, or you can specify a
                                             raster. This technique works well when 
                                             all of the rasters have a similar histogram.
                                    
                                             Standard_Deviation-Change each of the pixel's
                                             values according to its relationship with the
                                             histogram of the target raster, within one 
                                             standard deviation. The standard deviation can be
                                             calculated from all of the rasters in the mosaic
                                             dataset, or you can specify a target raster. 
                                             This technique works best when all of the 
                                             rasters have normal distributions.
    ------------------------------------     --------------------------------------------------------------------
    dodging_surface_type                     Required string.When using the Dodging balance method, 
                                             each pixel needs a target color, which is determined by 
                                             the surface type.

                                             Single_Color-Use when there are only a small
                                             number of raster datasets and a few different
                                             types of ground objects. If there are too many
                                             raster datasets or too many types of ground 
                                             surfaces, the output color may become blurred. 
                                             All the pixels are altered toward a single 
                                             color point-the average of all pixels. 
                                    
                                             Color_Grid- Use when you have a large number
                                             of raster datasets, or areas with a large 
                                             number of diverse ground objects. Pixels 
                                             are altered toward multiple target colors, 
                                             which are distributed across the mosaic dataset. 

                                             First_Order- This technique tends to create a 
                                             smoother color change and uses less storage in
                                             the auxiliary table, but it may take longer to
                                             process compared to the color grid surface. 
                                             All pixels are altered toward many points obtained
                                             from the two-dimensional polynomial slanted plane. 

                                             Second_Order-This technique tends to create a 
                                             smoother color change and uses less storage in
                                             the auxiliary table, but it may take longer to
                                             process compared to the color grid surface. 
                                             All input pixels are altered toward a set of 
                                             multiple points obtained from the two-dimensional 
                                             polynomial parabolic surface. 

                                             Third_Order-This technique tends to create a 
                                             smoother color change and uses less storage in
                                             the auxiliary table, but it may take longer to
                                             process compared to the color grid surface. 
                                             All input pixels are altered toward multiple 
                                             points obtained from the cubic surface.
    ------------------------------------     --------------------------------------------------------------------
    target_image                             Optional. The image service you want to use to color balance 
                                             the images in the image collection.
                                             It can be a portal Item or an image service URL or a URI
    ------------------------------------     --------------------------------------------------------------------
    context                                  Optional dictionary. It contains additional settings that allows
                                             users to customize the statistics computation settings.

                                             Example:
                                             {"skipRows": 10, "skipCols": 10, "reCalculateStats": "OVERWRITE"}
    ------------------------------------     --------------------------------------------------------------------
    gis                                      Optional GIS. the GIS on which this tool runs. If not specified, the active GIS is used.
    ====================================     ====================================================================

    :return:
        The imagery layer url

    '''

    gis = arcgis.env.active_gis if gis is None else gis

    params = {}
    context = {}
    _set_image_collection_param(gis,params, image_collection)

    color_correction_allowed_values = ['DODGING', 'HISTOGRAM', 'STANDARD_DEVIATION']

    if not color_correction_method.upper() in color_correction_allowed_values:
            raise RuntimeError('similarity can only be one of the following: '+ str(color_correction_allowed_values))        
    params['colorCorrectionMethod'] = color_correction_method

    dodging_surface_type_allowed_values = ['SINGLE_COLOR', 'COLOR_GRID', 'FIRST_ORDER','SECOND_ORDER','THIRD_ORDER']
    if not dodging_surface_type.upper() in dodging_surface_type_allowed_values:
            raise RuntimeError('similarity can only be one of the following: '+ str(dodging_surface_type_allowed_values))
    params['dodgingSurface'] = dodging_surface_type

    if target_image is not None:
        if isinstance(target_image, str):
            if 'http:' in target_image or 'https:' in target_image:
                params['targetImage'] = json.dumps({ 'url' : target_image })
            else:
                params['targetImage'] = json.dumps({ 'uri' : target_image })
        elif isinstance(target_image, Item):
                params['targetImage'] = json.dumps({ "itemId" : target_image.itemid })
        else:
            raise TypeError("target_image should be a string (url or uri) or Item")
    
    _set_context(params, context)        
    
    task = 'ComputeColorCorrection'
    job_values = _execute_task(gis, task, params)


    return job_values["result"]["url"]



###################################################################################################
## Compute Control Points
###################################################################################################
def compute_control_points(image_collection, reference_image=None, image_location_accuracy="High", context = None, gis=None):
    '''
    This service tool is used for computing matching control points between images
    within an image collection and/or matching control points between the image 
    collection images and the reference image.
    http://pro.arcgis.com/en/pro-app/tool-reference/data-management/compute-control-points.htm
    
    ====================================    ====================================================================
    **Argument**                            **Description**
    ------------------------------------    --------------------------------------------------------------------
    image_collection                        Required. This is the image collection that will be adjusted.

                                            The image_collection can be a portal Item or an image service URL or a URI
                            
                                            The image_collection must exist.
    ------------------------------------    --------------------------------------------------------------------
    reference_image                         This is the reference image service that can be used to generate ground control 
                                            points set with the image service. 
                                            It can be a portal Item or an image service URL or a URI
    ------------------------------------    --------------------------------------------------------------------
    image_location_accuracy                 Optional string. This option allows users to specify the location accuracy of the  
                                            imagery.
                                            VERYLOW, LOW, MEDIUM, HIGH

                                            LOW-Images have a large shift and a large rotation (> 5 degrees).
                                                The SIFT algorithm will be used in the point matching computation. 

                                            MEDIUM-Images have a medium shift and a small rotation (<5 degrees).
                                                   The Harris algorithm will be used in the point matching computation. 

                                            HIGH-Images have a small shift and a small rotation.
                                                 The Harris algorithm will be used in the point matching computation. 

                                            Default is HIGH
    ------------------------------------    --------------------------------------------------------------------
    context                                 Optional dictionary. Context contains additional environment settings that affect 
                                            output control points generation.
                                            {
                                            "pointSimilarity":"MEDIUM",
                                            "pointDensity": "MEDIUM",
                                            "pointDistribution": "RANDOM"
                                            }
    ------------------------------------    --------------------------------------------------------------------
    gis                                     Optional GIS. the GIS on which this tool runs. If not specified, the active GIS is used.
    ====================================    ====================================================================

    :return:
        The imagery layer url

    '''
    gis = arcgis.env.active_gis if gis is None else gis

    params = {}
    context = {}
    _set_image_collection_param(gis, params, image_collection)

    if reference_image is not None:
        if isinstance(reference_image, str):
            if 'http:' in reference_image or 'https' in reference_image:
                params['referenceImage'] = json.dumps({ 'url' : reference_image })
            else:
                params['referenceImage'] = json.dumps({ 'uri' : reference_image })
        elif isinstance(reference_image, Item):
                params['referenceImage'] = json.dumps({ "itemId" : reference_image.itemid })
        else:
            raise TypeError("reference_image should be a string (url or uri) or Item")

    params["imageLocationAccuracy"]=image_location_accuracy

    _set_context(params, context)  

    task = 'ComputeControlPoints'
    job_values = _execute_task(gis, task, params)

    return job_values["result"]

###################################################################################################
## Compute Seamlines
###################################################################################################
def compute_seamlines(image_collection,
                      seamlines_method,
                      context = None,
                      gis = None):
    '''
    Compute seamlines on the image collection. This service tool is used to compute
    seamlines for the image collection, usually after the image collection has been
    block adjusted. Seamlines are helpful for generating the seamless mosaicked 
    display of overlapped images in image collection. The seamlines are computed
    only for candidates that will eventually be used for generating the result
    ortho-mosaicked image.  
    http://pro.arcgis.com/en/pro-app/tool-reference/data-management/build-seamlines.htm

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    image_collection       Required, the input image collection that will be adjusted.
                           The image_collection can be a portal Item or an image service URL or a URI
                           The image_collection must exist.
    ------------------     --------------------------------------------------------------------
    seamlines_method       Required string. These are supported methods for generated seamlines for the image collection.
    
                            VORONOI-Generate seamlines using the area Voronoi diagram. 

                            DISPARITY-Generate seamlines based on the disparity images of stereo pairs.
                            
                            GEOMETRY - Generate seamlines for overlapping areas based on the intersection 
                            of footprints. Areas with no overlapping imagery will merge the footprints. 

                            RADIOMETRY - Generate seamlines based on the spectral patterns of features 
                            within the imagery.

                            EDGE_DETECTION - Generate seamlines over intersecting areas based on the 
                            edges of features in the area.

                            This method can avoid seamlines cutting through buildings. 
    ------------------     --------------------------------------------------------------------
    context                Optional dictionary. Context contains additional settings that allows users to customize
                           the seamlines generation. 
                           Example:
                           {"minRegionSize": 100,
                           "pixelSize": "",
                           "blendType": "Both",
                           "blendWidth": null,
                           "blendUnit": "Pixels",
                           "requestSizeType": "Pixels",
                           "requestSize": 1000,
                           "minThinnessRatio": 0.05,
                           "maxSilverSize": 20
                           }

                           Allowed keys are:
                           "minRegionSize", "pixelSize", "blendType", "blendWidth", 
                           "blendUnit", "requestSizeType", "requestSize", 
                           "minThinnessRatio", "maxSilverSize"
    ------------------     --------------------------------------------------------------------
    gis                    Optional GIS. The GIS on which this tool runs. If not specified, the active GIS is used.
    ==================     ====================================================================

    :return:
        The Imagery layer url

    '''
    gis = arcgis.env.active_gis if gis is None else gis

    params = {}
    context={}
    _set_image_collection_param(gis, params, image_collection)

    contextAllowedValues= {"minRegionSize", "pixelSize", "blendType", "blendWidth", 
                           "blendUnit", "requestSizeType", "requestSize", 
                           "minThinnessRatio", "maxSilverSize"
                           }

    seamlines_method_allowed_values = ['VORONOI', 'DISPARITY','GEOMETRY', 'RADIOMETRY', 'EDGE_DETECTION']
    if not seamlines_method.upper() in seamlines_method_allowed_values:
            raise RuntimeError('similarity can only be one of the following: '+str(seamlines_method_allowed_values))
    params['seamlinesMethod'] = seamlines_method

    _set_context(params, context)

    task = 'ComputeSeamlines'
    job_values = _execute_task(gis, task, params)

    return job_values["result"]["url"]


###################################################################################################
## Edit control points
###################################################################################################
def edit_control_points(image_collection, control_points, gis = None):
    '''
    This service can be used to append additional ground control point sets to
    the image collection's control points. It is recommended that a ground control point (GCP) set 
    should contain one ground control point and multiple tie points. 
    The service tool can also be used to edit tie point sets. 
    The input control points dictionary will always replace the points in the tie points
    table if the point IDs already exist. 
   
    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    image_collection       Required.
                           The image_collection can be a portal Item or an image service URL or a URI
                           The image_collection must exist.
    ------------------     --------------------------------------------------------------------
    control_points         Required, a list of control point sets objects.

                           The schema of control points follows the schema 
                           of the mosaic dataset control point table. 

                           The control point object should contain the point geometry, pointID, type, status and the
                           imagePoints. (the imagePoints attribute inside the control points object lists the imageIDs)

                           -- pointID (int) - The ID of the point within the control point table.

                           -- type (int)    - The type of the control point as determined by its numeric value
                                                 1: Tie Point 
                                                 2: Ground Control Point.
                                                 3: Check Point

                           -- status (int)  - The status of the point. A value of 0 indicates that the point will
                                                 not be used in computation. A non-zero value indicates otherwise.

                           

                           -- imageID (int) - Image identification using the ObjectID from the mosaic dataset footprint table.

                           Example:
                           [{
                           "status": 1,
                           "type": 2,
                           "x": -117.0926538,
                           "y": 34.00704253,
                           "z": 634.2175,
                           "spatialReference": {
                               "wkid": 4326
                           }, // default WGS84
                           "imagePointSpatialReference": {}, // default ICS
                           "pointId": 1,
                           "xyAccuracy": "0.008602325",
                           "zAccuracy": "0.015",
                           "imagePoints": [{
                               "imageID": 1,
                               "x": 2986.5435987557084,
                               "y": -2042.5193648409431,
                               "u": 3057.4580682832734,
                               "v": -1909.1506872159698
                           },
                           {
                               "imageID": 2,
                               "x": 1838.2814361401108,
                               "y": -2594.5280063817972,
                               "u": 3059.4079724863363,
                               "v": -2961.292545463305
                           },
                           {
                               "imageID": 12,
                               "x": 5332.855578204663,
                               "y": -2533.2805429751907,
                               "u": 614.2338676573158,
                               "v": -165.10836768947297
                           },
                           {
                               "imageID": 13,
                               "x": 4932.0895715254455,
                               "y": -1833.8401744114287,
                               "u": 616.9396928182223,
                               "v": -1243.1445126959693
                           }]
                           },
                           …
                           …
                           ] 


    ------------------     --------------------------------------------------------------------
    gis                    Optional GIS. The GIS on which this tool runs. If not specified, the active GIS is used.
    ==================     ====================================================================

    :return:
        The Imagery layer url

    '''
    gis = arcgis.env.active_gis if gis is None else gis

    params = {}

    _set_image_collection_param(gis, params, image_collection)

    params['inputControlPoints'] = json.dumps(control_points)

    task = 'EditControlPoints'
    job_values = _execute_task(gis, task, params)

    return job_values["result"]["url"]

###################################################################################################
## Generate DEM
###################################################################################################
def generate_dem(image_collection,
                 out_dem,
                 cell_size,
                 surface_type,
                 matching_method = None,
                 context = None,
                 gis = None):
    '''
    Generate a DEM from the image collection. Refer to "Interpolate From Point Cloud"
    GP tool for more documentation
    http://pro.arcgis.com/en/pro-app/tool-reference/data-management/interpolate-from-point-cloud.htm
    
    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    image_collection       Required. The input image collection that will be used
                           to generate the DEM from.
                           The image_collection can be a portal Item or an image service URL or a URI
                           The image_collection must exist.
    ------------------     --------------------------------------------------------------------
    out_dem                This is the output digital elevation model image url or name. 
                           Like Raster Analysis services, the service can be an existing multi-tenant service URL 
                           or this service tool can create it. 
                           It can be a url, uri, portal item, string representing the name of output image service 
    ------------------     --------------------------------------------------------------------
    cell_size              Required, The cell size of the output raster dataset. This is a single numeric input. 
                           Rectangular cell size such as {"x": 10, "y": 10} is not supported. 
                           The cell size unit will be the unit used by the image collection's spatial reference.
    ------------------     --------------------------------------------------------------------
    surface_type           Required string. Create a digital terrain model or a digital surface model. Refer
                           to "surface_type" parameter of the GP tool.
                           The available choices are:

                           DTM - Digital Terrain Model, the elevation is only the elevation of the bare earth, not including structures above the surface.

                           DSM - Digital Surface Model, the elevation includes the structures above the surface, for example, buildings, trees, bridges.
    ------------------     --------------------------------------------------------------------
    matching_method        Optional string. The method used to generate 3D points. 

                           ETM-A feature-based stereo matching that uses the Harris operator to 
                           detect feature points. It is recommended for DTM generation.  

                           SGM- Produces more points and more detail than the ETM method. It is 
                           suitable for generating a DSM for urban areas. This is more 
                           computationally intensive than the ETM method1.  

                           MVM (Multi-view image matching (MVM) - is based on the SGM matching method followed by a fusion step in which 
                           the redundant depth estimations across single stereo model are merged. 
                           It produces dense 3D points and is computationally efficient

                           References:  
                           Heiko Hirschmuller et al., "Memory Efficient Semi-Global Matching," 
                           ISPRS Annals of the Photogrammetry, Remote Sensing and Spatial 
                           Information Sciences, Volume 1-3, (2012): 371-376. 

                           Refer to the documentation
                           of "matching_method" parameter of the "Generate Point Cloud"
                           GP tool @ http://pro.arcgis.com/en/pro-app/tool-reference/data-management/generate-point-cloud.htm
    ------------------     --------------------------------------------------------------------
    context                Optional dictionary. Additional allowed point cloud generation parameter and DEM 
                           interpolation parameter can be assigned here.  
                           For example: 
                           Point cloud generation parameters -  
                           {"maxObjectSize": 50, 
                           "groundSpacing": None, 
                           "minAngle": 10, 
                           "maxAngle": 70, 
                           "minOverlap": 0.6, 
                           "maxOmegaPhiDif": 8, 
                           "maxGSDDif": 2, 
                           "numImagePairs": 2, 
                           "adjQualityThreshold": 0.2, 
                           "regenPointCloud": False 
                           } 

                           DEM interpolation parameters -  
                           {"method": "TRIANGULATION", 
                           "smoothingMethod": "GAUSS5x5", 
                           "applyToOrtho": True, 
                           "fillDEM": "https://...." 
                           } 
 
                           Note:  
                           The "applyToOrtho" flag can apply the generated DEM back into the 
                           mosaic dataset's geometric function to achieve more accurate 
                           orthorectification result.  
                           The "fillDEM" flag allows the user to specify an elevation service URL as 
                           background elevation to fill the area when elevation model pixels cannot be 
                           interpolated from the point cloud.  
    ------------------     --------------------------------------------------------------------
    gis                    Optional GIS. The GIS on which this tool runs. If not specified, the active GIS is used.
    ==================     ====================================================================

    :return:
        The DEM layer item

    '''

    gis = arcgis.env.active_gis if gis is None else gis

    task = 'GenerateDEM'

    contextAllowedValues= ["maxObjectSize", "groundSpacing", "minAngle", "maxAngle", "minOverlap", "maxOmegaPhiDif", 
                            "maxGSDDif", "numImagePairs", "adjQualityThreshold", "method", "smoothingMethod", "applyToOrtho"]
    context = {}
    params = {}

    _set_image_collection_param(gis, params, image_collection)

    if isinstance(out_dem, Item):
        params["outputDEM"] = json.dumps({"itemId": out_dem.itemid})
    elif isinstance(out_dem, str):
        if ("/") in out_dem or ("\\") in out_dem:
            if 'http:' in out_dem or 'https:' in out_dem:
                params['outputDEM'] = json.dumps({ 'url' : out_dem })
            else:
                params['outputDEM'] = json.dumps({ 'uri' : out_dem })
        else:
            result = gis.content.search("title:"+str(out_dem), item_type = "Imagery Layer")
            if len(result)>0:
                result = result[0]
                if result is not None:
                    params["outputDEM"]= json.dumps({"itemId": result.itemid})
            else:
                doesnotexist = gis.content.is_service_name_available(out_dem, "Image Service") 
                if doesnotexist:
                    params["outputDEM"] = json.dumps({"serviceProperties": {"name" : out_dem}})
              
    
    params['cellSize'] = cell_size

    surfaceTypeAllowedValues = ['DTM', 'DSM']
    if not surface_type in surfaceTypeAllowedValues:
        raise RuntimeError('surface_type can only be one of the following: '+ str(surfaceTypeAllowedValues))
    params['surfaceType'] = surface_type
    
    if matching_method is not None:
        params['matchingMethod'] = matching_method

    _set_context(params, context)  
    
    job_values = _execute_task(gis, task, params)

    output_service= gis.content.get(job_values["result"]["itemId"])

    return  output_service 


###################################################################################################
## Generate orthomosaic
###################################################################################################
def generate_orthomosaic(image_collection,
                         out_ortho,
                         regen_seamlines=True,
                         recompute_color_correction=True,
                         context=None,
                         gis=None):
    '''
    Function can be used for generating single ortho-rectified mosaicked image from image collection after 
    the block adjustment.  
    
    ===================================    ====================================================================
    **Argument**                           **Description**
    -----------------------------------    --------------------------------------------------------------------
    image_collection                       Required. The input image collection that will be used
                                           to generate the ortho-mosaic from.
                                           The image_collection can be a portal Item or an image service URL or a URI
                                           The image_collection must exist.
    -----------------------------------    --------------------------------------------------------------------
    out_ortho                               Required. This is the output ortho-mosaicked image converted from the image 
                                            collection after the block adjustment.   
                                            It can be a url, uri, portal item, string representing the name of output image 
                                            service  
    -----------------------------------    --------------------------------------------------------------------
    regen_seamlines                        Optional, boolean. 

                                           Choose whether to apply seamlines before the orthomosaic image generation or not. 
                                           The seamlines will always be regenerated if this parameter is set to True. 
                                           The user can set the seamline options through the context parameter. 
                                           If the seamline generation options are not set, the default will be used.  

                                            Default value is True
    -----------------------------------    --------------------------------------------------------------------
    recompute_color_correction              Optional, boolean.
                                            Choose whether to apply color correction settings to the output ortho-image or not. 
                                            Color correction will always be recomputed if this option is set to True. 
                                            The user can configure the compute color correction settings through the context parameter. 
                                            If there is no color collection setting, the default will be used.  

                                            Default value is True
    -----------------------------------    --------------------------------------------------------------------
    context                                Optional dictionary. Context contains additional environment settings that affect output 
                                           image. The supported environment settings for this tool are:

                                           1. Output Spatial Reference (outSR)-the output features will
                                              be projected into the output spatial reference.

                                           2. Extent (extent) - extent that would clip or expand the output image 
    -----------------------------------    --------------------------------------------------------------------
    gis                                    Optional GIS. The GIS on which this tool runs. If not specified, the active GIS is used.
    ===================================    ====================================================================

    :return:
        The Orthomosaicked Imagery layer item

    '''
    gis = arcgis.env.active_gis if gis is None else gis

    task = 'GenerateOrthomosaic'

    params = {}

    _set_image_collection_param(gis, params, image_collection)
        
    if isinstance(out_ortho, Item):
        params["outputOrthoImage"] = json.dumps({"itemId": out_ortho.itemid})
    elif isinstance(out_ortho, str):
        if ("/") in out_ortho or ("\\") in out_ortho:
            if 'http:' in out_ortho or 'https:' in out_ortho:
                params['outputOrthoImage'] = json.dumps({ 'url' : out_ortho })
            else:
                params['outputOrthoImage'] = json.dumps({ 'uri' : out_ortho })
        else:
            result = gis.content.search("title:"+str(out_ortho), item_type = "Imagery Layer")
            if len(result) > 0:
                result = result[0]
                if result is not None:
                    params["outputOrthoImage"]= json.dumps({"itemId": result.itemid})
            else:
                doesnotexist = gis.content.is_service_name_available(out_ortho, "Image Service") 
                if doesnotexist:
                    params["outputOrthoImage"] = json.dumps({"serviceProperties": {"name" : out_ortho}})
            

    if regen_seamlines is not None:
        if not isinstance(regen_seamlines, bool):
            raise TypeError("The 'regen_seamlines' parameter must be a boolean")
        params['regenSeamlines'] = regen_seamlines

    if recompute_color_correction is not None:
        if not isinstance(recompute_color_correction, bool):
            raise TypeError("The 'recompute_color_correction' parameter must be a boolean")
        params['recomputeColorCorrection'] = recompute_color_correction    

    _set_context(params, context)   
    
    job_values = _execute_task(gis, task, params)

    output_service= gis.content.get(job_values["result"]["itemId"])

    return  output_service 


###################################################################################################
## Generate report
###################################################################################################
def generate_report(image_collection, gis = None):
    """
    This function is used to generate orthomapping report with image collection 
    that has been block adjusted. The report would contain information about 
    the quality of the adjusted images, the distribution of the control points, etc.
    The output of this service tool is a downloadable html page. 

    ===================    ====================================================================
    **Argument**           **Description**
    -------------------    --------------------------------------------------------------------
    image_collection       Required. the input image collection that should be
                           used to generate a report from.
                           The image_collection can be a portal Item or an image service URL or a URI
                           The image_collection must exist.
    -------------------    --------------------------------------------------------------------
    gis                    Optional GIS. The GIS on which this tool runs. If not specified, the active GIS is used.
    ===================    ====================================================================

    :return:
        The URL of a single html webpage that is a formatted orthomapping report

    """
    gis = arcgis.env.active_gis if gis is None else gis

    params = {}

    _set_image_collection_param(gis, params, image_collection)

    task = 'GenerateReport'
    job_values = _execute_task(gis, task, params)

    return job_values["outReport"]["url"]

###################################################################################################
## query camera info
###################################################################################################
def query_camera_info(camera_query=None,
                      gis = None):
    ''' 
    This service tool is used to query specific or the entire digital camera 
    database. The digital camera database contains the specs
    of digital camera sensors that were used to capture drone images. 

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    camera_query           Required String. This is a SQL query statement that can 
                           be used to filter a portion of the digital camera
                           database.
    ------------------     --------------------------------------------------------------------
    gis                    Optional GIS. The GIS on which this tool runs. If not specified, the active GIS is used.
    ==================     ====================================================================


    :return:
        Data Frame representing the camera database

    '''
    import pandas as pd
    import numpy as np

    gis = arcgis.env.active_gis if gis is None else gis

    params = {}

    if camera_query is not None:
        if not isinstance(camera_query, str):
            raise TypeError("The 'camera_query' parameter must be a string")
        params['query'] = camera_query
   
    task = 'QueryCameraInfo'
    job_values = _execute_task(gis, task, params)
    pd.set_option('display.max_rows', None)
    df = pd.DataFrame(np.array(job_values["outputCameraInfo"]["content"]),columns = ["Maker","Model", "Focal Length", "Columns", "Rows","Pixel Size"])
    display(df)
    return df
    #return job_values["outputCameraInfo"]["content"]

###################################################################################################
## query control points
###################################################################################################
def query_control_points(image_collection,
                         query,
                         gis = None):
    '''
    Query for control points in an image collection. It allows users to query 
    among certain control point sets that has ground control points inside.

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    image_collection       Required, the input image collection on which to query 
                           the the control points.

                           The image_collection can be a portal Item or an image service URL or a URI.
                           
                           The image_collection must exist.
    ------------------     --------------------------------------------------------------------
    query                  Required string. a SQL statement used for querying the point;

                           e.g. "pointID > 100"
    ------------------     --------------------------------------------------------------------
    gis                    Optional GIS. The GIS on which this tool runs. If not specified, the active GIS is used.
    ==================     ====================================================================


    :return:
        A dictionary object

    '''
    gis = arcgis.env.active_gis if gis is None else gis

    params = {}

    _set_image_collection_param(gis, params, image_collection)

    if not isinstance(query, str):
        raise TypeError("The 'query' parameter must be a string")

    params['where'] = query

    task = 'QueryControlPoints'
    job_values = _execute_task(gis, task, params)

    if job_values["outControlPoints"] is not None:
        gptool_url = gis.properties.helperServices.orthoMapping.url
        gptool = arcgis.gis._GISResource(gptool_url, gis)
        result = gptool._con.post(job_values["outControlPoints"]["url"],{},token=gptool._token)
    else:
        return job_values["outControlPoints"]

    return result

###################################################################################################
## Reset image collection
###################################################################################################
def reset_image_collection(image_collection,
                           gis = None):
    '''
    Reset the image collection. It is used to reset the image collection to its 
    original state. The image collection could be adjusted during the orthomapping 
    workflow and if the user is not satisfied with the result, they will be able 
    to clear any existing adjustment settings and revert the images back to 
    un-adjusted state

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    image_collection       Required, the input image collection to reset
                           The image_collection can be a portal Item or an image service URL or a URI.
                           
                           The image_collection must exist.
    ------------------     --------------------------------------------------------------------
    gis                    Optional GIS. The GIS on which this tool runs. If not specified, the active GIS is used.
    ==================     ====================================================================

    :return:
        A boolean indicating whether the reset was successful or not

    '''
    gis = arcgis.env.active_gis if gis is None else gis

    params = {}

    _set_image_collection_param(gis, params, image_collection)

    task = 'ResetImageCollection'
    job_values = _execute_task(gis, task, params)
    return job_values["result"]


def compute_spatial_reference_factory_code(latitude, longitude): 
    """
    Computes spatial reference factory code. This value may be used as out_sr value in create image collection function

    Parameters
    ----------
    latitude : latitude value in decimal degress that will be used to compute UTM zone
    longitude : longitude value in decimal degress that will be used to compute UTM zone

    Returns
    -------
    factory_code : spatial reference factory code
    """
    from math import isnan, fabs, floor
    zone = 0
    if (isnan(longitude) or isnan(latitude) or fabs(longitude) > 180.0 or fabs(latitude) > 90.0):
        raise RuntimeError("Incorrect latitude or longitude value")

    zone = floor((longitude + 180)/6) + 1
    if (latitude >= 56.0 and latitude < 64.0 and longitude >= 3.0 and longitude < 12.0):
        zone = 32;

    if (latitude >= 72.0 and latitude < 84.0):
        if  (longitude >= 0.0  and longitude <  9.0):
            zone = 31;
        elif (longitude >= 9.0  and longitude < 21.0):
            zone = 33;
        elif (longitude >= 21.0 and longitude < 33.0):
            zone = 35;
        elif (longitude >= 33.0 and longitude < 42.0): 
            zone = 37

    if(latitude>=0):
        srid = 32601
    else:
        srid = 32701

    factory_code = srid + zone -1

    return factory_code



"""
This resource provides information about the data holdings of the
server. This information is used by ArcGIS for Desktop and other
clients to validate data paths referenced by GIS services.
You can register new data items with the server by using the
Register Data Item operation. Use the Find Data Items operation to
search through the hierarchy of data items.
The Compute Ref Count operation counts and lists all references to a
specific data item. This operation helps you determine if a
particular data item can be safely deleted or refreshed.
"""
from __future__ import absolute_import
import os
import re
import json
from .._common import BaseServer
from .._common.util import contextmanager, _tempinput
########################################################################
class DataStoreManager(BaseServer):
    """
    This resource provides information about the data holdings of the
    server, as well as the ability to manage (add new items, update primary
    data store, remove a data store item, etc) the data store. Data items
    are used by ArcGIS for Desktop and other clients to validate data paths
    referenced by GIS services.

    .. note::
        A relational data store type represents a database platform that has been
        registered for use on a portal's hosting server by the ArcGIS Server
        administrator. Each relational data store type describes the
        properties ArcGIS Server requires in order to connect to an instance of
        a database for a particular platform. At least one registered
        relational data store type is required before client applications such
        as Insights for ArcGIS can create Relational Database Connection portal
        items.


    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    url                    Required string. The URL to the Data Store URL.
    ------------------     --------------------------------------------------------------------
    gis                    Optional string. The GIS, Server, or ServicesDirectory object.
    ==================     ====================================================================

    """
    _con = None
    _json_dict = None
    _url = None
    _json = None
    _gis = None
    _datastores = None
    #----------------------------------------------------------------------
    def __init__(self,
                 url,
                 gis=None,
                 **kwargs):
        """Constructor
            Inputs:
               url - admin url
               gis - gis object
               initialize - optional initializes the componenents in the class
        """
        connection = kwargs.pop('connection', None)
        initialize = False
        super(DataStoreManager, self).__init__(
            gis=gis,
            url=url)
        if hasattr(gis, '_con'):
            self._con = gis._con
        elif hasattr(gis, 'post'):
            self._con = gis
        if connection:
            self._con = connection
        self._url = url
        if initialize:
            self._init()
    #----------------------------------------------------------------------
    def __str__(self):
        return '<%s for %s>' % (type(self).__name__, self._url)
    #----------------------------------------------------------------------
    def __repr__(self):
        return '<%s for %s>' % (type(self).__name__, self._url)
    #----------------------------------------------------------------------
    def list(self):
        """Retrieves a list of datastore objects.

        :return:
           The list of datastore items.
        """
        self._datastores = None
        if self._datastores is None:
            self._datastores = []
            for item in self.data_items['rootItems']:
                for path in self.search(parent_path=item)['items']:
                    self._datastores.append(Datastore(datastore=self,
                                                      path=path['path'],
                                                      datadict=None))
        return self._datastores
    #----------------------------------------------------------------------
    @property
    def config(self):
        """
           Gets the data store configuration properties. These properties
           affect the behavior of the data holdings of the server. For
           example, the blockDataCopy property - when this property is false,
           or not set at all, copying data to the site when publishing services
           from a client application is allowed. This is the default behavior.
           When this property is true, the client application is not allowed to
           copy data to the site when publishing. Rather, the publisher is
           required to register data items through which the service being
           published can reference data. Values: true | false
        """

        """ jenn note -- need link or list of the possible data store configuration properties."""
        params = {
            "f" : "json"
        }
        url = self._url + "/config"
        return self._con.get(path=url, params=params)
    #----------------------------------------------------------------------
    @config.setter
    def config(self, config):
        """
        This operation allows you to update the data store configuration
        You can use this to allow or block the automatic copying of data
        to the server at publish time

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        config                 Required string. A JSON string containing the data store configuration.
        ==================     ====================================================================

        :return:
           JSON dictionary of the set configuration properties.

        """
        if config is None:
            config = {}
        params = {
            "f" : "json",
            "datastoreConfig" : config
        }
        url = self._url + "/config/update"
        return self._con.post(path=url, postdata=params)
    #----------------------------------------------------------------------
    def get(self, path):
        """
        Retrieves the data item object at the given path.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        path                   Required string. The path to the data item.
        ==================     ====================================================================


        :return:
            The data item object, None if not found.
        """
        if path[0] != "/":
            path = "/%s" % path
        params = {"f" : "json"}
        urlpath = self._url + "/items" + path

        datadict = self._con.post(urlpath, params)
        if 'status' not in datadict:
            return Datastore(self, "/items" + path, datadict)
        else:
            return None
    #----------------------------------------------------------------------
    def add_folder(self,
                   name,
                   server_path,
                   client_path=None):
        """
        Registers a folder with the data store.


        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        name                   Required string. The unique fileshare name on the server.
        ------------------     --------------------------------------------------------------------
        server_path            Required string. The path to the folder from the server (and client, if shared path).
        ------------------     --------------------------------------------------------------------
        client_path            Optional string. If folder is replicated, the path to the folder from
                               the client; if folder is shared, don't set this parameter.
        ==================     ====================================================================

        :return:
            The data item if successfully registered, None otherwise.

        """
        conn_type = "shared"
        if client_path is not None:
            conn_type = "replicated"

        item = {
            "type" : "folder",
            "path" : "/fileShares/" + name,
            "info" : {
                "path" : server_path,
                "dataStoreConnectionType" : conn_type
            }
        }
        if client_path is not None:
            item['clientPath'] = client_path
        res = self._register_data_item(item=item)
        if res['status'] == 'success' or res['status'] == 'exists':
            return Datastore(self, "/fileShares/" + name)
        else:
            return None
        return
    #----------------------------------------------------------------------
    def add(self,
            name,
            item):
        """
        Registers a new data item with the data store.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        name                   Required string. The name of the new data item.
        ------------------     --------------------------------------------------------------------
        item                   Required string. The dictionary representing the data item.
                               See http://resources.arcgis.com/en/help/arcgis-rest-api/index.html#//02r3000001s9000000
        ==================     ====================================================================


        :return:
            The data item if registered successfully, None otherwise.
        """
        res = self._register_data_item(item=item)
        if res['status'] == 'success' or res['status'] == 'exists':

            return Datastore(self, item['path'])
        else:
            #print(str(res))
            return None
    #----------------------------------------------------------------------
    def add_bigdata(self,
                    name,
                    server_path=None,
                    connection_type="fileShare"):
        """
        Registers a bigdata fileshare with the data store.


        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        name                Required string. The unique bigdata fileshare name on the server.
        ---------------     --------------------------------------------------------------------
        server_path         Optional string. The path to the folder from the server.
        ---------------     --------------------------------------------------------------------
        connection_type     Optional string. Allows for the setting of the types of big data store.
                            The value 'fileShare' is used for local big data stores, and for
                            cloud stores, the connection_type should be 'dataStore'. The value
                            'fileShare' is the default value.
        ===============     ====================================================================

        :return:
           The big data fileshare if registered successfully, None otherwise.
        """
        output = None
        pattern = r'\\\\[a-zA-Z]+'
        if re.match(pattern, server_path) is not None:  # starts with double backslash, double the backslashes
            server_path = server_path.replace('\\', '\\\\')

        path_str = '{"path":"' + server_path + '"}'

        item = {
                "path": "/bigDataFileShares/" + name,
                "type": "bigDataFileShare",

                "info": {
                    "connectionString": path_str,
                    "connectionType": connection_type
                }
            }
        res = self._register_data_item(item=item)

        if res['status'] == 'success' or res['status'] == 'exists':
            output = Datastore(self, "/bigDataFileShares/" + name)
        return output
    #----------------------------------------------------------------------
    def generate_connection_string(self, sde):
        """
        Converts an SDE connection file to a string with encrypted password.


        ===============   ====================================================
        **Parameters**    **Description**
        ---------------   ----------------------------------------------------
        sde               required string.  Path to SDE connection file.
        ===============   ====================================================

        returns: string on success, None on failure

        **Usage**:

        >>> con = dm.create_connection_string(r"c:\myfolder\postgres_db.sde")
        >>> print(con)
        'ENCRYPTED_PASSWORD=************************;SERVER=localhost;
        INSTANCE=sde:postgresql:localhost,5432;DBCLIENT=postgresql;
        DB_CONNECTION_PROPERTIES=localhost,5432;DATABASE=esri_spatial;
        USER=sde;VERSION=sde.DEFAULT;AUTHENTICATION_MODE=DBMS'

        """
        if str(sde).lower().endswith('.sde'):
            from arcgis.gis.server._common import ServerConnection
            from arcgis.gis.server.catalog import ServicesDirectory
            from arcgis.gis.server import Uploads

            up = Uploads(url=self._con.baseurl.replace("rest/services", "admin/uploads"),
                         gis=self._con)

            if self._con.portal_connection:
                d = ServicesDirectory(url=self._con.baseurl,
                                      portal_connection=self._con.portal_connection)
            elif isinstance(self._con, ServerConnection):
                d = ServicesDirectory(url=self._con.baseurl)
                d._con = self._con

            try:
                service = d.get("PublishingTools", 'System')
            except:
                service = d.get("PublishingToolsEx", 'System')
            upload_res = up.upload(path=sde)
            if upload_res[0] == True:
                res = service.get_database_connection_string(
                    in_conndatatype="UPLOADED_CONNECTION_FILE_ID",
                    in_inputdata=upload_res[1]['item']['itemID'])
                up.delete(item_id=upload_res[1]['item']['itemID'])
                return res
        return
    #----------------------------------------------------------------------
    def add_cloudstore(self, name, conn_str, object_store,
                       provider, managed=False, folder=None):
        """
        Cloud Store data item represents a connection to a Amazon or Microsoft Azure store.
        Connection information for the data store item is stored within conn_str as a
        stringified JSON. ArcGIS Server encrypts connection string for storage. Connection
        strings that are encrypted will include a {crypt} prefix. You can get a data store
        item with decrypted connection string by passing a decrypt=true parameter in the request
        for a data store item. Data store with decrypted connection string will be returned only for
        requests made with https. The examples below show data stores with decrypted conn_str.
        A valid object_store (S3 bucket or Azure Blob store) is required. Folders within an object
        store are optional.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        name                Required string. The name of the cloud store.
        ---------------     --------------------------------------------------------------------
        conn_str            Required string. The connection information for the cloud storage
                            product.
        ---------------     --------------------------------------------------------------------
        object_store        Required string. This is the amazon bucket path or Azuze path.
        ---------------     --------------------------------------------------------------------
        provider            Required string. Values must be amazon or azure.
        ---------------     --------------------------------------------------------------------
        managed             Optional boolean. When the data store is server only, the database
                            is entirely managed and owned by the server and cannot be accessed
                            by the publisher directly. When this option is chosen, the
                            managed property should be set to true. Otherwise it is false.
        ---------------     --------------------------------------------------------------------
        folder              Optional string. For some Azure cloud stores, an optional folder
                            can be specified.
        ===============     ====================================================================


        :return: DataStore

        """
        item = {
            "path": "/cloudStores/%s" % name,
            "type": "cloudStore",
            "provider": provider,
            "info": {
                "isManaged": managed,
                "connectionString": conn_str,
                "objectStore": object_store
            }
        }
        if folder is not None:
            item['info']['folder'] = folder
        res = self._register_data_item(item=item)
        if res['status'] == 'success' or res['status'] == 'exists':
            return Datastore(self, "/cloudStores/" + name)
        else:
            return None
        return

    #----------------------------------------------------------------------
    def add_database(self,
                     name,
                     conn_str,
                     client_conn_str=None,
                     conn_type="shared"):
        """
        Registers a database with the data store.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        name                   Required string. The unique database name on the server.
        ------------------     --------------------------------------------------------------------
        conn_str               Required string. The path to the folder from the server (and client,
                               if shared or serverOnly database)
        ------------------     --------------------------------------------------------------------
        client_conn_str        Optional string. The connection string for client to connect to replicated enterprise database>
        ------------------     --------------------------------------------------------------------
        conn_type              Optional string. The connection type.  Default value is shared,
                               other choices are replicated or serverOnly
        ==================     ====================================================================


        :return:
            The data item if successfully registered, None otherwise.
        """

        item = {
            "type" : "egdb",
            "path" : "/enterpriseDatabases/" + name,
            "info" : {
                "connectionString" : conn_str,
                "dataStoreConnectionType" : conn_type
            }
        }

        if client_conn_str is not None:
            item['info']['clientConnectionString'] = client_conn_str

        is_managed = False
        if conn_type == "serverOnly":
            is_managed = True

        item['info']['isManaged'] = is_managed
        res = self._register_data_item(item=item)
        if res['status'] == 'success' or res['status'] == 'exists':
            return Datastore(self, "/enterpriseDatabases/" + name)
        else:
            return None
        return
    #----------------------------------------------------------------------
    def get_total_refcount(self, path):
        """
        The total number of references to a given data item
        that exists on the server. You can use this operation to
        determine if a data resource can be safely deleted, or taken
        down for maintenance.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        path                   Required string. The complete hierarchical path to the item.
        ==================     ====================================================================


        :return:
            A JSON dictionary containing a number representing the total count.

        """
        url = self._url + "/computeTotalRefCount"
        params = {
            "f" : "json",
            "path" : path
        }
        return self._con.post(path=url,
                              postdata=params)
    #----------------------------------------------------------------------
    def make_datastore_machine_primary(self,
                                       item_name,
                                       machine_name):
        """
        Promotes a standby machine to the primary Data Store machine. The
        existing primary machine is downgraded to a standby machine.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        item_name              Required string. The primary machine item name in the data store.
        ------------------     --------------------------------------------------------------------
        machine_name           Required string. The machine name of the machine to promote to primary.
        ==================     ====================================================================


        :return:
            A boolean indicating success (True) or failure (False).

        """
        url = self._url + "/items/enterpriseDatabases" + \
            "/{datastoreitem}/machines/{machine_name}/makePrimary".format(
                datastoreitem=item_name,
                machine_name=machine_name)
        params = {"f" : "json"}

        return self._con.post(path=url,
                              postdata=params)
    #----------------------------------------------------------------------
    def get_relational_datastore_type(self, type_id):
        """
        This resource lists the properties of a registered relational data
        store of the given type. The properties returned are those that client
        applications must provide when creating a Relational Database
        Connection portal item.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        type_id                Required string. The datastore type ID of interest.
                               See http://resources.arcgis.com/en/help/arcgis-rest-api/index.html#/Relational_Data_Store_Types/02r300000303000000/
        ==================     ====================================================================


        :return:
           A JSON string listing the properties

        """
        """ jenn note: useful would be a list of possible datastore type IDs -- ??? esri.teradata, esri.sqlserver, esri.hana """

        params = {"f" : "json"}
        url = self._url + "/relationalDatastoreTypes/{i}".format(
            i=type_id)
        return self._con.get(path=url,
                             params=params)
    #----------------------------------------------------------------------
    @property
    def relational_datastore_types(self):
        """
        Gets a list of the relational data store types that have been
        registered with the server. Each registered relational data store
        type has both an id and a name property, as well as an array of
        userDefinedProperties, which indicates the properties client
        applications must provide when creating a Relational Database
        Connection portal item. Only administrators can register and
        unregister a relational data store type. The following database
        platforms are supported: SAP HANA, Microsoft SQL Server and
        Teradata.
        """
        params = {"f" : "json"}
        url = self._url + "/relationalDatastoreTypes"
        return self._con.get(path=url,
                             params=params)
    #----------------------------------------------------------------------
    def search(self,
               parent_path=None,
               ancestor_path=None,
               types=None,
               id=None):
        """
        Use this operation to search through the various data items that are registered in the server's data store.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        parent_path            Optional string. The path of the parent under which to find items.
        ------------------     --------------------------------------------------------------------
        ancestor_path          Optional string. The path of the ancestor under which to find items.
        ------------------     --------------------------------------------------------------------
        types                  Optional string. A filter for the type of the items (for example, fgdb or folder or egdb).
        ------------------     --------------------------------------------------------------------
        id                     Optional string. A filter to search by the ID of the item.
        ==================     ====================================================================


        :return:
            A JSON list of the items found matching the search criteria.
        """

        """ jenn note: list of possible types """
        params = {
            "f" : "json",
        }
        if parent_path is not None:
            params['parentPath'] = parent_path
        if ancestor_path is not None:
            params['ancestorPath'] = ancestor_path
        if types is not None:
            params['types'] = types
        if id is not None:
            params['id'] = id
        url = self._url + "/findItems"
        return self._con.post(path=url,
                              postdata=params)
    #----------------------------------------------------------------------
    def _register_data_item(self, item):
        """
        Registers a new data item with the server's data store.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        item                   Required string. The JSON representing the data item.
                               See http://resources.arcgis.com/en/help/arcgis-rest-api/index.html#//02r3000001s9000000
        ==================     ====================================================================

        :return:
            A response
        """
        params = {
            "item" : item,
            "f" : "json"
        }
        url = self._url + "/registerItem"
        return self._con.post(path=url,
                              postdata=params)
    #----------------------------------------------------------------------
    @property
    def data_items(self):
        """
        Gets the list of data items that are the root of all other data items in the data store.
        """
        url = self._url + "/items"
        params = {
            "f" : "json"
        }
        return self._con.get(path=url,
                             params=params)
    #----------------------------------------------------------------------
    def validate(self):
        """
        Validates all the items in the data store.

        :return:
            True if all items are valid.
        """
        params = {
            "f" : "json"
        }
        url = self._url + "/validateAllDataItems"
        return self._con.post(path=url, postdata=params)
    #----------------------------------------------------------------------
    def make_primary(self, datastore_name, machine_name):
        """
        Promotes a standby machine to the primary Data Store machine. The
        existing primary machine is downgraded to a standby machine.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        datastore_name         Required string. The primary machine name in the data store.
        ------------------     --------------------------------------------------------------------
        machine_name           Required string. The machine name of the machine to promote to primary.
        ==================     ====================================================================


        :return:
            A boolean indicating success (True) or failure (False).

        """
        url = self._url + "/items/enterpriseDatabases/%s/machines/%s/makePrimary" % (datastore_name, machine_name)
        params = {
            "f" : "json"
        }
        return self._con.post(path=url, postdata=params)
    #----------------------------------------------------------------------
    def remove_datastore_machine(self, item_name, machine_name):
        """
        Removes a standby machine from the Data Store. This operation is
        not supported on the primary Data Store machine.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        item_name              Required string. The standby machine item name in the data store.
        ------------------     --------------------------------------------------------------------
        machine_name           Required string. The machine name of the machine to remove.
        ==================     ====================================================================

        :return:
            A boolean indicating success (True) or failure (False).

        """
        url = self._url + "/items/enterpriseDatabases/%s/machines/%s/remove" % (item_name, machine_name)
        params = {
            "f" : "json"
        }
        return self._con.post(path=url, postdata=params)
    #----------------------------------------------------------------------
    def start(self, item_name, machine_name):
        """
        Starts the database instance running on the Data Store machine.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        item_name              Required string. The database item name in the data store to start.
        ------------------     --------------------------------------------------------------------
        machine_name           Required string. The machine name of the machine with the database
                               instance to start.
        ==================     ====================================================================

        :return:
           A boolean indicating success (True) or failure (False).

        """
        url = self._url + "/items/enterpriseDatabases/%s/machines/%s/start" % (item_name, machine_name)
        params = {
            "f": "json"
        }
        return self._con.post(path=url, postdata=params)
    #----------------------------------------------------------------------
    def stop(self, item_name, machine_name):
        """
        Stop the database instance running on the Data Store machine.


        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        item_name              Required string. The database item name in the data store to stop.
        ------------------     --------------------------------------------------------------------
        machine_name           Required string. The machine name of the machine with the database
                               instance to stop.
        ==================     ====================================================================

        :return:
           A boolean indicating success (True) or failure (False).

        """
        url = self._url + "/items/enterpriseDatabases/%s/machines/%s/stop" % (item_name,
                                                                              machine_name)
        params = {
            "f": "json"
        }
        return self._con.post(path=url, postdata=params)
    #----------------------------------------------------------------------
    def _unregister_data_item(self, path):
        """
        Unregisters a data item that has been previously registered with
        the server's data store.


        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        path                   Required string. The path to the share folder.
        ==================     ====================================================================

        :return:
            T

            .. code-block:: python

            EXAMPLE:

            path = r"/fileShares/folder_share"
            print data.unregisterDataItem(path)

        """
        url = self._url + "/unregisterItem"
        params = {
            "f" : "json",
            "itempath" : path
        }
        return self._con.post(path=url, postdata=params)
    #----------------------------------------------------------------------
    def validate_egdb(self, data_store_name, name):
        """
        Checks the status of the given ArcGIS Data Store and provides a health check response.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        data_store_name        Required string. The item name of the data store.
        ------------------     --------------------------------------------------------------------
        name                   Required string. The machine name of where the data store is.
        ==================     ====================================================================

        :return:
            A JSON response containing general status information and an overall health report.

        """
        url = self._url + "/items/enterpriseDatabases/%s/machines/%s/validate" % (data_store_name,
                                                                                  name)
        params = {
            "f" : "json"
        }
        return self._con.post(path=url, postdata=params)
###########################################################################
class Datastore(BaseServer):
    """
    Represents a single Datastore in the Data Store Manager.


    """
    _path = None
    _datastore = None
    _json_dict = None
    _json = None
    _con = None
    _url = None
    def __init__(self, datastore, path, datadict=None, **kwargs):
        self._path = path

        super(Datastore, self).__init__(datastore=datastore,
                                        path=path,
                                        url=datastore._url + "%s" % path,
                                        connection=datastore._con,
                                        initialize=True,
                                        datadict=datadict)
        path = "/items%s" % path
        if datastore:
            self._con = datastore._con
        self._datastore = datastore
        self._url = "%s%s" % (datastore._url, path)
        self._init()
    #----------------------------------------------------------------------
    def __str__(self):
        state = ["   %s=%r" % (attribute, value) for (attribute, value) in self._json_dict.items()]
        return '\n'.join(state)
    #----------------------------------------------------------------------
    def __repr__(self):
        return '<%s title:"%s" type:"%s">' % (type(self).__name__, self._url, self.type)
    #----------------------------------------------------------------------
    @property
    def manifest(self):
        """
        Gets the manifest resource for a big data file share.
        """
        data_item_manifest_url = self._url + "/manifest"
        if data_item_manifest_url.find('/bigDataFileShares') != -1:
            params = {
                'f': 'json',
            }
            res = self._con.post(data_item_manifest_url,
                                 params,
                                 verify_cert=False)
        else:
            res = {}
        return res
    #----------------------------------------------------------------------
    @manifest.setter
    def manifest(self, value):
        """
        Sets the manifest resource for a big data file share.
        """
        manifest_upload_url =  self._url + '/manifest/update'
        if manifest_upload_url.find('/bigDataFileShares') != -1:
            with _tempinput(json.dumps(value)) as tempfilename:
                # Build the files list (tuples)
                files = []
                files.append(('manifest', tempfilename, os.path.basename(tempfilename)))

                postdata = {
                    'f' : 'pjson'
                }

                resp = self._.con.post(manifest_upload_url, postdata, files, verify_cert=False)
                if resp['status'] == 'success':
                    return True
                else:
                    return False
        else:
            return None
    #---------------------------------------------------------------------
    @property
    def hints(self):
        """
        Gets the hints resource for a big data file share. Hints
        are advanced parameters to control the generation of a manifest.
        """
        params = {
            'download' : True,
            'read' : True
        }
        url = self._url + "/hints"
        return self._con.get(path=url,
                             params=params)
    #---------------------------------------------------------------------
    @hints.setter
    def hints(self,
              hints):
        """
        Sets the hints resource for a big data file share. Hints
        are advanced parameters to control the generation of a manifest.

        Upload a hints file for a big data file share item. This will
        replace the existing hints file. To apply the control parameters in
        the hints file and regenerate the manifest, use the editDataItem to
        edit the big data file share (using the same data store item as
        input) which will regenerate the manifest. When a manifest is
        regenerated, it will be updated only for datasets that have hints
        and for new datasets that are added to the existing big data file
        share location.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        name                   Required string. The name of the big data item to update.
        ------------------     --------------------------------------------------------------------
        hints                  Required string. The hints file to be uploaded.
        ==================     ====================================================================

        """
        params = {
            "f" : "json"
        }
        files = {"hints" : hints}

        url = self._url + "/hints/update"
        if url.find('/bigDataFileShares') == -1:
            return None
        return self._con.post(path=url,
                              files=files,
                              postdata=params)
    #----------------------------------------------------------------------
    @property
    def ref_count(self):
        """
        The total number of references to this data item that exist on the
        server. You can use this property to determine if this data item
        can be safely deleted or taken down for maintenance.

        :return:
           A number indictaing the number of references to this data item.
        """
        return self.totalRefCount
    #----------------------------------------------------------------------
    def delete(self):
        """
        Unregisters this data item from the data store.

        :return:
           A boolean indicating success (True) or failure (False).
        """
        params = {
            "f" : "json" ,
            "itempath" : self.path,
            "force": True
        }
        path = self._datastore._url + "/unregisterItem"

        resp = self._con.post(path, params, verify_cert=False)
        if resp:
            return resp.get('success')
        else:
            return False
    #----------------------------------------------------------------------
    def update(self, item):
        """
        Edits this data item to update its connection information.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        item                   Required string. The dict representation of the updated item.
        ==================     ====================================================================

        :return:
            True if the data item was successfully updated, False if the update failed.

        """
        params = {
            "f" : "json" ,
            "item" : item
        }
        path = self._datastore._url +  "/items" + self.path +  "/edit"

        resp = self._con.post(path, params, verify_cert=False)
        if resp ['status'] == 'success':
            return True
        else:
            return False
    #----------------------------------------------------------------------
    def validate(self):
        """
        Validates that this data item's path (for file shares) or connection string (for databases)
        is accessible to every server node in the site. This is necessary for the data item to be
        registered and used successfully with the server's data store.

        :return:
            True if the data item was successfully validated.
        """
        params = {
            "f": "json",
            "item": self._json_dict
        }
        path = self._datastore._url + "/validateDataItem"
        if 'provider' in params['item'] and \
           params['item']['provider'] == 'ArcGIS Data Store':
            path = self._url + "/machines/" + params['item']['info']['machines'][0]['name'] + "/validate"
            res = self._con.post(path, {"f": "json"}, verify_cert=False)
        else:
            res = self._con.post(path, params, verify_cert=False)

        return res['status'] == 'success'
    #----------------------------------------------------------------------
    @property
    def datasets(self):
        """
        Gets the datasets in the data store (currently implemented for big data file shares).
        """
        data_item_manifest_url = self._url + "/manifest"

        params = {
            'f': 'json'
        }

        try:
            res = self._con.post(data_item_manifest_url, params, verify_cert=False)
            return res['datasets']
        except:
            return None



"""
Classes to manage a GIS Collaboration
"""
from .. import GIS, Group
import functools

def _lazy_property(fn):
    '''Decorator that makes a property lazy-evaluated.
    '''
    # http://stevenloria.com/lazy-evaluated-properties-in-python/
    attr_name = '_lazy_' + fn.__name__

    @property
    @functools.wraps(fn)
    def _lazy_property(self):
        if not hasattr(self, attr_name):
            setattr(self, attr_name, fn(self))
        return getattr(self, attr_name)
    return _lazy_property

class CollaborationManager(object):
    _gis = None
    _basepath = None
    _pid = None
    def __init__(self, gis,portal_id=None):
        self._gis = gis
        self._portal = gis._portal
        self._pid = portal_id
        if portal_id is None:
            res = self._portal.con.get("portals/self")
            if 'id' in res:
                self._pid = res['id']
            else:
                raise Exception("Could not find the portal's ID")
        self._basepath = "portals/%s" % self._pid

    #----------------------------------------------------------------------
    def create(self,
               name,
               description,
               workspace_name,
               workspace_description,
               portal_group_id,
               host_contact_first_name,
               host_contact_last_name,
               host_contact_email_address,
               access_mode="sendAndReceive"):
        """
        The create method creates a collaboration. The host
        of the collaboration is the portal where it is created. The initial
        workspace for the collaboration is also created. A portal group in
        the host portal is linked to the workspace. The access mode for the
        host portal is set. The contact information associated with the
        host can be specified; otherwise, the contact information for the
        administrator user performing the operation will be used.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        name                            Required string. Name of the collaboration
        ---------------------------     --------------------------------------------------------------------
        description                     Required string. Description of the collaboration
        ---------------------------     --------------------------------------------------------------------
        workspace_name                  Required string. The name of the initial workspace.
        ---------------------------     --------------------------------------------------------------------
        workspace_description           Required string. The description of the initial workspace.
        ---------------------------     --------------------------------------------------------------------
        portal_group_id                 Required string. ID of group in the portal that will be linked with
                                        the workspace.
        ---------------------------     --------------------------------------------------------------------
        host_contact_first_name         Required string. The first name of the contact person for the
                                        collaboration host portal.
        ---------------------------     --------------------------------------------------------------------
        host_contact_last_name          Required string. The last name of the contact person for the
                                        collaboration host portal.
        ---------------------------     --------------------------------------------------------------------
        host_contact_email_address      Required string. The email address of the contact person for the
                                        collaboration host portal.
        ---------------------------     --------------------------------------------------------------------
        access_mode                     Required string. The organization's access mode to the workspace.
                                        Values: send | receive | sendAndReceive (default)
        ===========================     ====================================================================


        :return: the data item is registered successfully, None otherwise

        """
        if access_mode not in ['send', 'receive', 'sendAndReceive']:
            raise Exception("Invalid access_mode. Must be of value: send, " + \
                            "receive or sendAndReceive.")
        params = {
            "f" : "json",
            "name" : name,
            "description" : description,
            "workspaceName" : workspace_name,
            "workspaceDescription" : workspace_description,
            "portalGroupId" : portal_group_id,
            "hostContactFirstname" : host_contact_first_name,
            "hostContactLastname" : host_contact_last_name,
            "hostContactEmailAddress" : host_contact_email_address,
            "accessMode" : access_mode,
            "config" : {}
        }
        data_path = "%s/createCollaboration" % self._basepath
        res = self._portal.con.post(data_path, params)
        if 'collaboration' in res and \
           'id' in res['collaboration']:
            return Collaboration(collab_manager=self,
                                 collab_id=res['collaboration']['id'],
                                 portal_id=self._pid)
    #----------------------------------------------------------------------
    def accept_invitation(self,
                          first_name,
                          last_name,
                          email,
                          invitation_file=None,
                          invitation_JSON=None,
                          webauth_username=None,
                          webauth_password=None,
                          webauth_cert_file=None,
                          webauth_cert_password=None
                          ):
        """
        The accept_invitation operation allows a portal to accept a
        collaboration invitation. The invitation file received securely
        from the collaboration host portal must be provided. Once a guest
        accepts an invitation to a collaboration, it must link workspace(s)
        associated with the collaboration to local portal group(s). The
        guest must export a collaboration invitation response file and send
        it to the host. Once the host processes the response, content can
        be shared between the host and guest(s).

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        first_name                      Required string. The first name of the contact person for the guest
                                        portal.
        ---------------------------     --------------------------------------------------------------------
        last_name                       Required string. The last name of the contact person.
        ---------------------------     --------------------------------------------------------------------
        email                           Required string. The email of the contact person.
        ---------------------------     --------------------------------------------------------------------
        invitation_file                 Optional string. The invite file to upload to portal. Use either
                                        this parameter or invitation_JSON.
        ---------------------------     --------------------------------------------------------------------
        invitation_JSON                 Optional string. The same contents as the invitation_file parameter
                                        but passed as a string. Use either this parameter or invitation_file.
        ---------------------------     --------------------------------------------------------------------
        webauth_username                Optional string. If the collaboration host requires web-tier
                                        authentication, optionally use this parameter to provide the host's
                                        web-tier authentication user name.
        ---------------------------     --------------------------------------------------------------------
        webauth_password                Optional string. If the collaboration host requires web-tier
                                        authentication, optionally use this parameter to provide the host's
                                        web-tier authentication password.
        ---------------------------     --------------------------------------------------------------------
        webauth_cert_file               Optional string. If the collaboration host requires web-tier
                                        authentication, optionally use this parameter to provide the host's
                                        web-tier authentication certificate file.
        ---------------------------     --------------------------------------------------------------------
        webauth_cert_password           Optional string. If the collaboration host requires web-tier
                                        authentication, optionally use this parameter to provide the host's
                                        web-tier authentication certificate password.
        ===========================     ====================================================================

        :returns: dict

        """
        data_path = "%s/acceptCollaborationInvitation" % self._basepath
        params = {
            'f' : 'json',
            'guestContactFirstname' : first_name,
            'guestContactLastname' : last_name,
            'guestContactEmailAddress' : email
        }
        files =  None
        if invitation_file is None and \
           invitation_JSON is None:
            raise ValueError("invitation_file or invitation_JSON must be provided")
        if invitation_file:
            files = {}
            files['invitationFile'] = invitation_file
        if invitation_JSON:
            params['invitationJSON'] = invitation_JSON
        if webauth_cert_file:
            if files is None:
                files = {}
            files['hostWebauthCertificateFile'] = webauth_cert_file
        if webauth_cert_password:
            params['hostWebauthCertPassword'] = webauth_cert_password
        if webauth_password and webauth_username:
            params['hostWebauthUsername'] = webauth_username
            params['hostWebauthPassword'] = webauth_password
        con = self._portal.con
        return con.post(path=data_path,
                        postdata=params,
                        files=files)
    #----------------------------------------------------------------------
    def list(self):
        """gets all collaborations for a portal"""
        data_path = "%s/collaborations" % self._basepath
        params = {"f" : "json",
                  "num":100,
                  'start' : 1}
        res = self._portal.con.get(data_path, params)
        collabs = []
        while len(res['collaborations']) > 0:
            for collab in res['collaborations']:
                collabs.append(Collaboration(collab_manager=self,
                                             collab_id=collab['id']))
            res = self._portal.con.get(data_path, params)
            params['start'] = res['nextStart']
            if res['nextStart'] == -1:
                return collabs
        return collabs
    #----------------------------------------------------------------------
    def validate_invitation(self,
                            first_name,
                            last_name,
                            email,
                            invitation_file=None,
                            invitation_JSON=None,
                            webauth_username=None,
                            webauth_password=None,
                            webauth_cert_file=None,
                            webauth_cert_password=None):
        """
        The validate_invitation method allows a portal to
        validate a collaboration invitation. The invitation file received
        securely from the collaboration host portal must be provided.
        Validation checks include checking that the invitation is for the
        intended recipient.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        first_name                      Required string. The first name of the contact person for the guest
                                        portal.
        ---------------------------     --------------------------------------------------------------------
        last_name                       Required string. The last name of the contact person.
        ---------------------------     --------------------------------------------------------------------
        email                           Required string. The email of the contact person.
        ---------------------------     --------------------------------------------------------------------
        invitation_file                 Optional string. The invite file to upload to portal. Use either
                                        this parameter or invitation_JSON.
        ---------------------------     --------------------------------------------------------------------
        invitation_JSON                 Optional string. The same contents as the invitation_file parameter
                                        but passed as a string. Use either this parameter or invitation_file.
        ---------------------------     --------------------------------------------------------------------
        webauth_username                Optional string. If the collaboration host requires web-tier
                                        authentication, optionally use this parameter to provide the host's
                                        web-tier authentication user name.
        ---------------------------     --------------------------------------------------------------------
        webauth_password                Optional string. If the collaboration host requires web-tier
                                        authentication, optionally use this parameter to provide the host's
                                        web-tier authentication password.
        ---------------------------     --------------------------------------------------------------------
        webauth_cert_file               Optional string. If the collaboration host requires web-tier
                                        authentication, optionally use this parameter to provide the host's
                                        web-tier authentication certificate file.
        ---------------------------     --------------------------------------------------------------------
        webauth_cert_password           Optional string. If the collaboration host requires web-tier
                                        authentication, optionally use this parameter to provide the host's
                                        web-tier authentication certificate password.
        ===========================     ====================================================================

        :returns: dict

        """
        data_path = "%s/validateCollaborationInvitation" % self._basepath
        params = {
            'f' : 'json',
            'guestContactFirstname' : first_name,
            'guestContactLastname' : last_name,
            'guestContactEmailAddress' : email
        }
        files =  None
        if invitation_file is None and \
           invitation_JSON is None:
            raise ValueError("invitation_file or invitation_JSON must be provided")
        if invitation_file:
            files = {}
            files['invitationFile'] = invitation_file
        if invitation_JSON:
            params['invitationJSON'] = invitation_JSON
        if webauth_cert_file:
            if files is None:
                files = {}
            files['hostWebauthCertificateFile'] = webauth_cert_file
        if webauth_cert_password:
            params['hostWebauthCertPassword'] = webauth_cert_password
        if webauth_password and webauth_username:
            params['hostWebauthUsername'] = webauth_username
            params['hostWebauthPassword'] = webauth_password
        con = self._portal.con
        return con.post(path=data_path,
                        postdata=params,
                        files=files)

    # ----------------------------------------------------------------------
    def collaborate_with(self, guest_gis, collaboration_name, collaboration_description):
        """
        A high level method to quickly establish a collaboration between two GIS. This method uses defaults
        wherever applicable and internally calls the `create`, `accept_invitation` and `invite_participant` methods.
        This method will create a new group and a new workspace in both the host and guest GIS for this collaboration.
        Invitation and response files created during the collaborations will be downloaded to the current working
        directory.

        **Use the other methods if you need fine-grained control over how the collaboration is set up.**

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        guest_gis                       Required GIS. GIS object of the guest org or Enterprise.
        ---------------------------     --------------------------------------------------------------------
        collaboration_name              Required string. A generic name for the collaboration. This name is
                                        used with prefixes such as wksp_<your_collab_name>,
                                        grp_<your_collab_name> to create the collaboration workspace and
                                        groups.
        ---------------------------     --------------------------------------------------------------------
        collaboration_description       Optional string. A generic description for the collaboration.
        ===========================     ====================================================================

        :returns: boolean



        """

        # create a group in the host
        host_group = self._gis.groups.create(title="grp_" + collaboration_name, tags='collaboration',
                                             description='Group for ' + collaboration_description)

        #create a collaboration in the host
        host_first_name = ""
        host_last_name = ""
        host_email = ""
        if hasattr(self._gis.users.me, 'firstName'):
            host_first_name = self._gis.users.me.firstName
            host_last_name = self._gis.users.me.lastName
        elif hasattr(self._gis.users.me, 'fullName'):
            sp = self._gis.users.me.fullName.split()
            host_first_name = sp[0]
            if len(sp) > 1:
                host_last_name = sp[1]
            else:
                host_last_name=host_first_name
        if hasattr(self._gis.users.me, 'email'):
            host_email = self._gis.users.me.email

        host_collab = self.create(name='collab_' + collaboration_name, description=collaboration_description,
                                  workspace_name='wksp_' + collaboration_name,
                                  workspace_description='Workspace for ' + collaboration_description,
                                  portal_group_id=host_group.id,
                                  host_contact_first_name=host_first_name,
                                  host_contact_last_name=host_last_name,
                                  host_contact_email_address=host_email)

        #Invite guest GIS as participant
        config = [{host_collab.workspaces[0]['id']:"sendAndReceive"}]
        invite_file = host_collab.invite_participant(config, guest_gis=guest_gis)

        #Create a group in guest GIS
        guest_group = guest_gis.groups.create(title='grp_' + collaboration_name, tags='collaboration',
                                              description='Group for ' + collaboration_description)

        #Accept invitation in guest GIS
        guest_first_name = ""
        guest_last_name = ""
        guest_email = ""
        if hasattr(guest_gis.users.me, 'firstName'):
            guest_first_name = guest_gis.users.me.firstName
            guest_last_name = guest_gis.users.me.lastName
        elif hasattr(guest_gis.users.me, 'fullName'):
            sp = self._gis.users.me.fullName.split()
            guest_first_name = sp[0]
            if len(sp) > 1:
                guest_last_name = sp[1]
            else:
                guest_last_name = guest_first_name
        if hasattr(guest_gis.users.me, 'email'):
            guest_email = guest_gis.users.me.email
        response = guest_gis.admin.collaborations.accept_invitation(first_name=guest_first_name,
                                                                    last_name=guest_last_name,
                                                                    email=guest_email,
                                                                    invitation_file=invite_file)

        #Export response from guest GIS
        guest_collab = None
        response_file = None
        if response['success']:
            guest_collab = Collaboration(guest_gis.admin.collaborations, host_collab.id)
            response_file = guest_collab.export_invitation('./')
        else:
            raise Exception("Unable to accept collaboration in the guest GIS")

        #Add guest group to guest collab
        group_add_result = guest_collab.add_group_to_workspace(guest_group, guest_collab.workspaces[0])

        #Accept response back in the host GIS
        host_collab.import_invitation_response(response_file)

        return True

###########################################################################
class Collaboration(dict):
    """
    The collaboration resource returns information about the collaboration
    with a specified ID.
    """
    _id = None
    _cm = None # CollaborationManager
    _baseurl = None
    _portal = None
    def __init__(self, collab_manager, collab_id, portal_id=None):
        dict.__init__(self)
        self._id = collab_id
        self._cm = collab_manager
        self._portal = collab_manager._gis._portal
        if portal_id is None:
            res = self._portal.con.get("portals/self")
            if 'id' in res:
                portal_id = res['id']
            else:
                raise Exception("Could not find the portal's ID")
        self._basepath = "portals/%s/collaborations/%s" % (portal_id,collab_id)
        params = {"f" : "json"}
        datadict = self._portal.con.post(self._basepath, params, verify_cert=False)

        if datadict:
            self.__dict__.update(datadict)
            super(Collaboration, self).update(datadict)
    #----------------------------------------------------------------------
    def __getattr__(self, name): # support group attributes as group.access, group.owner, group.phone etc
        try:
            return dict.__getitem__(self, name)
        except:
            raise AttributeError("'%s' object has no attribute '%s'" % (type(self).__name__, name))
    #----------------------------------------------------------------------
    def __getitem__(self, k): # support group attributes as dictionary keys on this object, eg. group['owner']
        try:
            return dict.__getitem__(self, k)
        except KeyError:
            params = { "f" : "json" }
            datadict = self._portal.con.post(self._basepath, params,
                                             verify_cert=False)
            super(Collaboration, self).update(datadict)
            self.__dict__.update(datadict)
            return dict.__getitem__(self, k)
    #----------------------------------------------------------------------
    def add_workspace(self, name, description, config, portal_group_id):
        """
        The add_workspace resource adds a new workspace to a
        portal-to-portal collaboration. Only collaboration hosts can create
        new workspaces.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        name                            Required string. The name of the workspace.
        ---------------------------     --------------------------------------------------------------------
        description                     Required string. Brief description of the workspace.
        ---------------------------     --------------------------------------------------------------------
        portal_group_id                 Required string. The ID of the portal group linked with the
                                        workspace.
        ===========================     ====================================================================

        :returns: dict

        """
        params = {
            "f": "json",
            "collaborationWorkspaceName" : name,
            "collaborationWorkspaceDescription" : description,
            "config" : config,
            "portalGroupId" : portal_group_id
        }
        path = "%s/%s" % (self._basepath, "addWorkspace")
        return self._portal.con.post(path, params, verify_cert=False)
    #----------------------------------------------------------------------
    def get_invitation(self, invitation_id):
        """
        The get_invitation operation returns the information about an
        invitation to participate in a portal-to-portal collaboration for a
        particular invitation with the specified ID.
        """
        params = {
            "f": "json"
        }
        path = "%s/%s/%s" % (self._basepath,
                             "invitations",
                             invitation_id)
        return self._portal.con.get(path, params)
    #----------------------------------------------------------------------
    def get_workspace(self, workspace_id):
        """
        The workspace resource provides information about the collaboration
        workspace with a specified ID.
        """
        params = {
            "f": "json"
        }
        path = "%s/%s/%s" % (self._basepath,
                             "workspaces",
                             workspace_id)
        return self._portal.con.get(path, params)
    #----------------------------------------------------------------------
    @property
    def invitations(self):
        """The invitations operation returns the invitation information for
        all the invitations generated by a portal-to-portal collaboration
        host.
        """
        params = {
            "f": "json",
            'start' : 1,
            'nun' : 100
        }
        invs = []
        path = "%s/%s" % (self._basepath, "invitations")
        res = self._portal.con.get(path, params)
        while len(res['collaborationInvitations']) > 0:
            invs += res['collaborationInvitations']
            params['start'] = res['nextStart']
            if res['nextStart'] == -1:
                return invs
            res = self._portal.con.get(path, params)
        return invs
    #----------------------------------------------------------------------
    def delete(self):
        """
        The delete operation deletes a portal-to-portal collaboration from
        the host portal. This stops any sharing set up from the
        collaboration. The collaboration will be removed on guest portals
        on the next refresh of their content based on the collaboration
        sharing schedule. Guests cannot delete collaborations, but they can
        discontinue participation in a collaboration via the
        removeParticipation endpoint.
        """
        params = {'f' : "json"}
        data_path = "%s/delete" % self._basepath
        resp = self._portal.con.post(data_path, params)
        if 'success' in resp:
            return resp['success']
        return resp
    #----------------------------------------------------------------------
    def remove_workspace(self, workspace_id):
        """
        The delete operation deletes a collaboration workspace. This
        immediately disables further replication of data to and from the
        portal and the collaboration participants.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        workspace_id                    Optional string. UID of the workspace to remove from the
                                        collaboration.
        ===========================     ====================================================================

        :returns: dict



        """
        params = {"f" : "json"}
        data_path = "%s/workpaces/%s/delete" % (self._basepath, workspace_id)
        return self._portal.con.post(data_path, params)
    #----------------------------------------------------------------------
    @_lazy_property
    def workspaces(self):
        """
        The workspaces resource lists all the workspaces in a given
        collaboration. A workspace is a virtual space in the collaboration
        to which each participating portal is either sending or receiving
        content. Workspaces can only be created by the collaboration owner.
        """
        data_path = "%s/workspaces" % self._basepath
        params = {'f' : 'json',
                  "num":100,
                  'start' : 1}
        res = self._portal.con.get(data_path, params)
        workspaces = []
        while len(res['workspaces']) > 0:
            workspaces += res['workspaces']
            params['start'] = res['nextStart']
            if res['nextStart'] == -1:
                return workspaces
            res = self._portal.con.get(data_path, params)
        return workspaces
    #----------------------------------------------------------------------
    def export_invitation(self, out_folder):
        """
        The exportInvitationResponse operation exports a collaboration
        invitation response file from a collaboration guest portal. The
        exported response file must be sent via email or through other
        communication channels that are established in your organization to
        the inviting portal's administrator. The inviting portal's
        administrator will then import your response file to complete the
        establishment of trust between your portals.
        It is important that the contents of this response file are not
        intercepted and tampered with by any unknown entity.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        out_folder                      Required string. Save location of the file.
        ===========================     ====================================================================

        :returns: string

        """
        params = {"f" : "json"}
        data_path = "%s/exportInvitationResponse" % self._basepath
        return self._portal.con.post(data_path, params, out_folder=out_folder, verify_cert=False)

    #----------------------------------------------------------------------
    def import_invitation_response(self,
                                   response_file,
                                   webauth_username=None,
                                   webauth_password=None,
                                   webauth_cert_file=None,
                                   webauth_cert_password=None):
        """
        The importInvitationResponse operation imports an invitation
        response file from a portal collaboration guest. The operation is
        performed on the portal that serves as the collaboration host. Once
        an invitation response is imported, trust between the host and the
        guest is established. Sharing of content between participants can
        proceed from this point.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        response_file                   Required string. File path to the response file.
        ---------------------------     --------------------------------------------------------------------
        webauth_username                Optional string. If the collaboration host requires web-tier
                                        authentication, optionally use this parameter to provide the host's
                                        web-tier authentication user name.
        ---------------------------     --------------------------------------------------------------------
        webauth_password                Optional string. If the collaboration host requires web-tier
                                        authentication, optionally use this parameter to provide the host's
                                        web-tier authentication password.
        ---------------------------     --------------------------------------------------------------------
        webauth_cert_file               Optional string. If the collaboration host requires web-tier
                                        authentication, optionally use this parameter to provide the host's
                                        web-tier authentication certificate file.
        ---------------------------     --------------------------------------------------------------------
        webauth_cert_password           Optional string. If the collaboration host requires web-tier
                                        authentication, optionally use this parameter to provide the host's
                                        web-tier authentication certificate password.
        ===========================     ====================================================================

        :returns: dict

        """
        params = {"f" : "json"}
        data_path = "%s/importInvitationResponse" % self._basepath
        files = {'invitationResponseFile' : response_file}
        if webauth_cert_file:
            files['guestWebauthCertificateFile'] = webauth_cert_file
        if webauth_cert_password:
            params['guestWebauthCertPassword'] = webauth_cert_password
        if webauth_username and webauth_password:
            params['guestWebauthUsername'] = webauth_username
            params['guestWebauthPassword'] = webauth_password
        con = self._portal.con
        return con.post(path=data_path,
                        postdata=params,
                        files=files,
                        verify_cert=False)
    #----------------------------------------------------------------------
    def invalidate(self, invitation_id):
        """
        The invalidate operation invalidates a previously generated
        portal-to-portal collaboration invitation. If a guest accepts this
        invitation and sends an invitation response for it, the response
        will not import successfully on the collaboration host.
        """
        params = {"f" : "json"}
        data_path = "%s/invitations/%s/invalidate" % (self._basepath, invitation_id)
        con = self._portal.con
        return con.post(data_path,
                       params, verify_cert=False)
    #----------------------------------------------------------------------
    def invite_participant(self,
                           config_json,
                           expiration=24,
                           guest_portal_url=None,
                           guest_gis=None,
                           save_path=None):
        """
        As a collaboration host, once you have set up a new collaboration,
        you are ready to invite other portals as participants in your
        collaboration. The inviteParticipant operation allows you to invite
        other portals to your collaboration by creating an invitation file.
        You need to send this invitation file to the administrator of the
        portal you are inviting to your collaboration. This can be done via
        email or through other communication channels that are established
        in your organization. It is important that the contents of this
        invitation file are not intercepted and tampered with by any
        unknown entity. The invitation file is in the format
        collaboration-<guestHostDomain>.invite.
        The administrator of the participant will accept the invitation by
        importing the invitation file into their portal. Their acceptance
        is returned to you as another file that you must import into your
        portal using the import_invitation_response operation. This will
        establish trust between your portal and that of your participant.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        config_json                     Required dict. A dict containing a map of access modes for the
                                        participant in each of the collaboration workspaces.
                                        Defined as: send | receive | sendAndReceive

                                        :Example:

                                        config_json = [
                                          {"workspace_id" : "send"},
                                          {"workspace_id2" : "receive"},
                                          {"workspace_id3" : "sendAndReceive"}
                                        ]
        ---------------------------     --------------------------------------------------------------------
        expiration                      Optional integer. The time in UTC when the invitation to collaborate
                                        should expire.
        ---------------------------     --------------------------------------------------------------------
        guest_portal_url                Optional string. The URL of the participating org or Enterprise that
                                        you want to invite to the collaboration.
        ---------------------------     --------------------------------------------------------------------
        guest_gis                       Optional GIS. GIS object to the guest collaboration site.
        ---------------------------     --------------------------------------------------------------------
        save_path                       Optional string. Path to download the invitation file to.
        ===========================     ====================================================================

        :returns: contents of a file that contains the invitation information


        """
        if guest_gis is None and \
           guest_portal_url is None:
            raise ValueError("A GIS object or URL is required")
        if guest_portal_url is None and \
           guest_gis:
            guest_portal_url = guest_gis._portal.url
        data_path = "%s/inviteParticipant" % self._basepath
        params = {
            "f" : "json",
            "guestPortalUrl" : guest_portal_url,
            "collaborationWorkspacesParticipantConfigJSON" : config_json,
            "expiration" : expiration
        }
        con = self._portal.con
        return con.post(path=data_path,
                       postdata=params,
                       verify_cert=False,
                        out_folder = save_path)
    #----------------------------------------------------------------------
    def get_participant(self, portal_id):
        """
        The participant operation provides information about the
        collaboration participant with a specified ID.
        """
        data_path = "%s/participants/%s" % (self._basepath, portal_id)
        params = {"f" : "json"}
        con = self._portal.con
        return con.get(data_path,
                       params)
    #----------------------------------------------------------------------
    def participants(self):
        """
        The participants resource provides information about all of the
        participants in a portal-to-portal collaboration.
        """
        data_path = "%s/participants" % self._basepath
        params = {"f" : "json"}
        con = self._portal.con
        return con.get(data_path,
                       params)
    #----------------------------------------------------------------------
    def add_group_to_workspace(self, portal_group, workspace):
        """
        This operation adds a group to a workspace that participates in a portal-to-portal collaboration. Content shared
         to the portal group is shared to other participants in the collaboration.


        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        portal_group                    Required Group of string. Group ID or object to add to the workspace.
        ===========================     ====================================================================


        :returns: dict

        """
        group_id = None
        if isinstance(portal_group, Group):
            group_id = portal_group.groupid
        elif isinstance(portal_group, str):
            group_id = portal_group

        data_path = "{}/workspaces/{}/updatePortalGroupLink".format(self._basepath, workspace['id'])
        params = {'f':'json',
                  'portalGroupId':group_id,
                  'enableRealtimeSync':True,
                  'copyFeatureServiceData': False}
        con = self._portal.con
        result = con.post(path=data_path, postdata=params, verify_cert=False)
        return result

    # ----------------------------------------------------------------------
    def _force_sync(self, workspace):
        """
        Undocumented. This operation will force sync the collaboration and its workspaces
        :param workspace:
        :return:
        """
        config_sync_data_path = "{}/configSync".format(self._basepath)
        config_sync_status = self._portal.con.get(config_sync_data_path, {"f":"json"})

        if config_sync_status['success']:
            #proceed to workspace sync
            workspace_sync_data_path = "{}/workspaces/{}/sync".format(self._basepath, workspace['id'])
            wksp_sync_status = self._portal.con.post(workspace_sync_data_path, postdata={'f':'json'}, verify_cert=False)
            return wksp_sync_status
        else:
            raise RuntimeError("Error force syncing")
    # ----------------------------------------------------------------------
    def refresh(self, invitation_id):
        """
        The refresh operation refreshes a previously generated
        portal-to-portal collaboration invitation. The new invitation file
        is provided via a multipart POST response. The expiration for the
        invitation is extended an additional 72 hours from the current
        time.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        invitation_id                   Required string. ID of the invitation to refresh
        ===========================     ====================================================================


        :returns: dict

        """
        params = {"f" : "json"}
        data_path = "%s/invitations/%s/refresh" % (self._basepath, invitation_id)
        con = self._portal.con
        return con.post(path=data_path,
                        postdata=params,
                       verify_cert=False)
    #----------------------------------------------------------------------
    def remove_participation(self):
        """
        The removeParticipation operation removes collaboration
        participation by a guest from a collaboration, allowing a guest to
        exit a collaboration. This immediately disables further
        replication of data to and from the portal and the other
        collaboration participants.
        """
        data_path = "%s/removeParticipation" % self._basepath
        params = {"f" : "json"}
        con = self._portal.con
        return con.post(path=data_path,
                       postdata=params,
                       verify_cert=False)
    #----------------------------------------------------------------------
    def remove_participant(self, portal_id):
        """
        The remove operation allows a collaboration host to remove a
        participant from a portal-to-portal collaboration.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        portal_id                       Required string. ID of the portal to remove.
        ===========================     ====================================================================


        :returns: dict

        """
        params = {'f' : 'json'}
        data_path = "%s/participants/%s/remove" % (self._basepath, portal_id)
        con = self._portal.con
        return con.post(path=data_path, postdata=params)
    #----------------------------------------------------------------------
    def remove_portal_group_link(self, workspace_id):
        """
        The remove_portal_group_link operation removes the link between a
        collaboration workspace and a portal group. Replication of content
        discontinues when the link is removed.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        workspace_id                    Required string. Workspace ID to remove from the link.
        ===========================     ====================================================================


        :returns: dict

        """
        params = {'f' : 'json'}
        data_path = "%s/workspaces/%s/removePortalGroupLink" % (self._basepath, workspace_id)
        con = self._portal.con
        return con.post(path=data_path, postdata=params)
    #----------------------------------------------------------------------
    def update_collaboration(self, name=None,
                             description=None, config=None):
        """
        The updateInfo operation updates certain properties of a
        collaboration, primarily its name, description, and configuration
        properties. The updates are propagated to guests when the next
        scheduled refresh of content occurs.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        name                            Optional string. Name of the collaboration
        ---------------------------     --------------------------------------------------------------------
        description                     Optional string. The description of the collaboration
        ---------------------------     --------------------------------------------------------------------
        config                          Optional dict. The configuration properties of the collaboration
        ===========================     ====================================================================

        :returns: dict

        """
        data_path = "%s/updateInfo" % self._basepath
        params = {"f" : "json"}
        if name:
            params['name'] = name
        if description:
            params['description'] = description
        if config:
            params['config'] = config
        con = self._portal.con
        return con.post(path=data_path, postdata=params)
    #----------------------------------------------------------------------
    def update_workspace(self,
                         workspace_id, name=None,
                         description=None, config=None):
        """
        The updateInfo operation updates certain collaboration workspace
        properties.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        workspace_id                    Required string. UID of the workspace
        ---------------------------     --------------------------------------------------------------------
        name                            Optional string. The name of the workspace
        ---------------------------     --------------------------------------------------------------------
        description                     Optional string. A brief set of texts that explains the workspace
        ---------------------------     --------------------------------------------------------------------
        config                          Optional dict. The configuration details of the new workspace
        ===========================     ====================================================================

        :returns: dict

        """
        data_path = "%s/workspaces/%s/updateInfo" % (self._basepath, workspace_id)
        params = {"f" : 'json'}
        if name:
            params['collaborationWorkspaceName'] = name
        if description:
            params['collaborationWorkspaceDescription'] = description
        if config:
            params['config'] = config
        con = self._portal.con
        return con.post(path=data_path, postdata=params)
    #----------------------------------------------------------------------
    def update_access_modes(self,
                            portal_id,
                            workspace_access_json):
        """
        The update_access_modes operation updates the access mode for a
        specific participant in a portal-to-portal collaboration.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        portal_id                       Required string. UID of the Portal
        ---------------------------     --------------------------------------------------------------------
        workspace_access_json           Required dict/string. JSON describing the participant's access mode.
        ===========================     ====================================================================

        :returns: dict

        """
        data_path = "/participants/%s/updateParticipantAccessModes" % portal_id
        params = {'f': 'json'}
        params['collaborationWorkspacesAccessJSON'] = workspace_access_json
        con = self._portal.con
        return con.post(path=data_path, postdata=params)
    #----------------------------------------------------------------------
    def update_portal_group_link(self, workspace_id,
                                 portal_id,
                                 enable_realtime_sync=True,
                                 interval_hours=1):
        """
        The updatePortalGroupLink operation updates the group linked with a
        workspace for a participant in a portal-to-portal collaboration.
        Content shared to the portal group is shared to other participants
        in the collaboration.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        workspace_id                    Required string. UID of the workspace
        ---------------------------     --------------------------------------------------------------------
        portal_id                       Required string. UID of the Portal
        ---------------------------     --------------------------------------------------------------------
        enable_realtime_sync            Optional boolean. Determines whether the content shared with the
                                        group is shared to other collaboration participants in real time,
                                        updating whenever changes are made, or whether the content is
                                        shared based on a schedule set by the collaboration host.
        ---------------------------     --------------------------------------------------------------------
        interval_hours                  Optional integer. sets the sharing schedule for the group
        ===========================     ====================================================================

        :returns: dict

        """
        data_path = "/workspaces/%s/updatePortalGroupLink" % workspace_id
        params = {
            'f': 'json',
            'portalGroupId' : portal_id,
            'enableRealtimeSync' : enable_realtime_sync,
            'syncIntervalHours' : interval_hours
        }

        con = self._portal.con
        return con.post(path=data_path, postdata=params)
    #----------------------------------------------------------------------
    def validate_invitation_response(self, response_file):
        """
        Prior to importing a collaboration invitation response, the
        invitation response file can be validated by using the
        validate_invitation_response operation to check for the existence
        of the collaboration and validity of the invitation response file.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        response_file                   Required string. Path to the collaboration response file.
        ===========================     ====================================================================

        :returns: dict

        """
        files = {'invitationResponseFile' : response_file}
        params = {'f':'json'}
        data_path = "%s/validatInvitationResponse" % self._basepath
        con = self._portal.con
        return con.post(path=data_path, postdata=params)


"""
The arcgis.raster module containing classes and raster analysis functions for working with raster data and
imagery layers.

Raster data is made up of a grid of cells, where each cell or pixel can have a value. Raster data is useful
for storing data that varies continuously, as in a satellite image, a surface of chemical concentrations, or
an elevation surface.

Use arcgis.raster.analytics.is_supported(gis) to check if raster analysis is supported in your GIS.
"""

from ._layer import ImageryLayer, RasterManager, ImageryTileManager, RasterCatalogItem
from .analytics import *
from .functions import *
from .orthomapping import *


